#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper classes for data schema dictionaries defined in the Scripting API.
"""

import six

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.utils.utils import merge_args_kwargs_dict
from wotpy.wot.enums import DataType


class DataSchemaDict(WotBaseDict):
    """Represents the common properties of a value type definition."""

    class Meta:
        fields = {
            "description",
            "title",
            "type",
            "const",
            "unit",
            "enum",
            "readOnly",
            "writeOnly"
        }

        defaults = {
            "readOnly": False,
            "writeOnly": False
        }

    @classmethod
    def build(cls, *args, **kwargs):
        """Builds an instance of the appropriate subclass for the given ValueType."""

        init_dict = merge_args_kwargs_dict(args, kwargs)

        klass_map = {
            DataType.NUMBER: NumberSchemaDict,
            DataType.BOOLEAN: BooleanSchemaDict,
            DataType.STRING: StringSchemaDict,
            DataType.OBJECT: ObjectSchemaDict,
            DataType.ARRAY: ArraySchemaDict,
            DataType.INTEGER: IntegerSchema
        }

        klass_type = init_dict.get("type")
        klass = klass_map.get(klass_type)

        if not klass:
            raise ValueError("Unknown type: {}".format(klass_type))

        return klass(*args, **kwargs)


class NumberSchemaDict(DataSchemaDict):
    """Properties to describe a numeric type."""

    class Meta:
        fields = DataSchemaDict.Meta.fields.union({
            "minimum",
            "maximum"
        })

        defaults = DataSchemaDict.Meta.defaults

    @property
    def type(self):
        """The type property represents the value type (a member of DataType)."""

        return DataType.NUMBER


class BooleanSchemaDict(DataSchemaDict):
    """Properties to describe a boolean type."""

    @property
    def type(self):
        """The type property represents the value type enumerated in DataType."""

        return DataType.BOOLEAN


class StringSchemaDict(DataSchemaDict):
    """Properties to describe a string type."""

    @property
    def type(self):
        """The type property represents the value type enumerated in DataType."""

        return DataType.STRING


class ObjectSchemaDict(DataSchemaDict):
    """Properties to describe an object type."""

    class Meta:
        fields = DataSchemaDict.Meta.fields.union({
            "properties",
            "required"
        })

        defaults = DataSchemaDict.Meta.defaults

    @property
    def type(self):
        """The type property represents the value type enumerated in DataType."""

        return DataType.OBJECT

    @property
    def properties(self):
        """Data schema nested definitions."""

        return {
            key: DataSchemaDict.build(val)
            for key, val in six.iteritems(self._init.get("properties", {}))
        }


class ArraySchemaDict(DataSchemaDict):
    """Properties to describe an array type."""

    class Meta:
        fields = DataSchemaDict.Meta.fields.union({
            "items",
            "minItems",
            "maxItems"
        })

        defaults = DataSchemaDict.Meta.defaults

    @property
    def type(self):
        """The type property represents the value type enumerated in DataType."""

        return DataType.ARRAY

    @property
    def items(self):
        """Used to define the characteristics of an array."""

        return DataSchemaDict.build(self._init["items"]) if "items" in self._init else None


class IntegerSchema(NumberSchemaDict):
    """Properties to describe an integer type."""

    @property
    def type(self):
        """The type property represents the value type enumerated in DataType."""

        return DataType.INTEGER



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Enumeration classes related to the HTTP server.
"""

from wotpy.utils.enums import EnumListMixin


class HTTPSchemes(EnumListMixin):
    """Enumeration of HTTP schemes."""

    HTTP = "http"
    HTTPS = "https"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import socket

import pytest
import tornado.gen
import tornado.ioloop
from aiozeroconf import ServiceStateChange, ServiceInfo
from faker import Faker
from six.moves import range

from tests.utils import find_free_port, run_test_coroutine
from wotpy.wot.discovery.dnssd.service import DNSSDDiscoveryService, build_servient_service_info
from wotpy.wot.servient import Servient


def _assert_service_added_removed(servient, service_history, instance_name=None):
    """Checks the service change history to assert that
    the servient service has been added and then removed."""

    info = build_servient_service_info(servient, instance_name=instance_name)
    servient_items = [item for item in service_history if item[1] == info.name]

    assert servient_items[-1][0] == service_history[-2][0] == DNSSDDiscoveryService.WOT_SERVICE_TYPE
    assert servient_items[-2][2] == ServiceStateChange.Added
    assert servient_items[-1][2] == ServiceStateChange.Removed


def _num_service_instance_items(servient, service_history, instance_name=None):
    """Returns the number of items in the given service history that match the servient."""

    info = build_servient_service_info(servient, instance_name=instance_name)
    return len([item for item in service_history if item[1] == info.name])


def test_start_stop():
    """The DNS-SD service can be started and stopped."""

    @tornado.gen.coroutine
    def test_coroutine():
        dnssd_discovery = DNSSDDiscoveryService()

        yield dnssd_discovery.start()

        assert dnssd_discovery.is_running

        for _ in range(10):
            yield dnssd_discovery.stop()

        assert not dnssd_discovery.is_running

        for _ in range(10):
            yield dnssd_discovery.start()

        assert dnssd_discovery.is_running

        yield dnssd_discovery.stop()

        assert not dnssd_discovery.is_running

    run_test_coroutine(test_coroutine)


def test_register(asyncio_zeroconf, dnssd_discovery):
    """WoT Servients may be registered for discovery on the DNS-SD service."""

    @tornado.gen.coroutine
    def test_coroutine():
        service_history = asyncio_zeroconf.pop("service_history")

        port_catalogue = find_free_port()
        servient = Servient(catalogue_port=port_catalogue)

        with pytest.raises(ValueError):
            yield dnssd_discovery.register(servient)

        yield dnssd_discovery.start()

        assert not len(service_history)

        yield dnssd_discovery.register(servient)

        while _num_service_instance_items(servient, service_history) < 1:
            yield tornado.gen.sleep(0.1)

        yield dnssd_discovery.stop()

        while _num_service_instance_items(servient, service_history) < 2:
            yield tornado.gen.sleep(0.1)

        _assert_service_added_removed(servient, service_history)

    run_test_coroutine(test_coroutine)


def test_unregister(asyncio_zeroconf, dnssd_discovery):
    """WoT Servients that have been previously registered
    on the DNS-SD service can be unregistered."""

    @tornado.gen.coroutine
    def test_coroutine():
        service_history = asyncio_zeroconf.pop("service_history")

        port_catalogue = find_free_port()
        servient = Servient(catalogue_port=port_catalogue)

        yield dnssd_discovery.start()
        yield dnssd_discovery.register(servient)
        yield dnssd_discovery.unregister(servient)

        while _num_service_instance_items(servient, service_history) < 2:
            yield tornado.gen.sleep(0.1)

        _assert_service_added_removed(servient, service_history)

    run_test_coroutine(test_coroutine)


def test_find(asyncio_zeroconf, dnssd_discovery):
    """Remote WoT Servients may be discovered using the DNS-SD service."""

    @tornado.gen.coroutine
    def test_coroutine():
        aio_zc = asyncio_zeroconf.pop("zeroconf")

        ipaddr = Faker().ipv4_private()
        port = find_free_port()
        service_name = "{}.{}".format(Faker().pystr(), DNSSDDiscoveryService.WOT_SERVICE_TYPE)
        server = "{}.local.".format(Faker().pystr())

        info = ServiceInfo(
            DNSSDDiscoveryService.WOT_SERVICE_TYPE,
            service_name,
            address=socket.inet_aton(ipaddr),
            port=port,
            properties={},
            server=server)

        yield aio_zc.register_service(info)

        with pytest.raises(ValueError):
            yield dnssd_discovery.find()

        yield dnssd_discovery.start()

        assert (ipaddr, port) in (yield dnssd_discovery.find(timeout=3))

    run_test_coroutine(test_coroutine)


def test_register_instance_name(asyncio_zeroconf, dnssd_discovery):
    """WoT Servients may be registered with custom service instance names."""

    @tornado.gen.coroutine
    def test_coroutine():
        service_history = asyncio_zeroconf.pop("service_history")

        port_catalogue = find_free_port()
        servient = Servient(catalogue_port=port_catalogue)

        instance_name = Faker().sentence()
        instance_name = instance_name.strip('.')[:32]

        yield dnssd_discovery.start()
        yield dnssd_discovery.register(servient, instance_name=instance_name)

        while _num_service_instance_items(servient, service_history, instance_name) < 1:
            yield tornado.gen.sleep(0.1)

        yield dnssd_discovery.stop()

        while _num_service_instance_items(servient, service_history, instance_name) < 2:
            yield tornado.gen.sleep(0.1)

        assert len([item[1].startswith(instance_name) for item in service_history]) == 2

        with pytest.raises(Exception):
            _assert_service_added_removed(servient, service_history)

        _assert_service_added_removed(servient, service_history, instance_name)

    run_test_coroutine(test_coroutine)


def test_enable_on_servient(asyncio_zeroconf, dnssd_servient):
    """The DNS-SD service may be enabled directly on the
    Servient to avoid the need of explicit instantiation."""

    @tornado.gen.coroutine
    def test_coroutine():
        service_history = asyncio_zeroconf.pop("service_history")
        instance_name = dnssd_servient.dnssd_instance_name

        yield dnssd_servient.start()

        while _num_service_instance_items(dnssd_servient, service_history, instance_name) < 1:
            yield tornado.gen.sleep(0.1)

        yield dnssd_servient.shutdown()

        while _num_service_instance_items(dnssd_servient, service_history, instance_name) < 2:
            yield tornado.gen.sleep(0.1)

        _assert_service_added_removed(dnssd_servient, service_history, instance_name)

    run_test_coroutine(test_coroutine)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import random
import uuid

import pytest
import tornado.gen
import tornado.ioloop
from faker import Faker

from wotpy.support import is_mqtt_supported
from wotpy.wot.dictionaries.interaction import ActionFragmentDict, EventFragmentDict, PropertyFragmentDict
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription
from wotpy.wot.thing import Thing

collect_ignore = []

if not is_mqtt_supported():
    logging.warning("Skipping MQTT tests due to unsupported platform")
    collect_ignore += ["test_server.py", "test_client.py"]


@pytest.fixture(params=[{"property_callback_ms": None}])
def mqtt_server(request):
    """Builds a MQTTServer instance that contains an ExposedThing."""

    from wotpy.protocols.mqtt.server import MQTTServer
    from tests.protocols.mqtt.broker import get_test_broker_url

    exposed_thing = ExposedThing(servient=Servient(), thing=Thing(id=uuid.uuid4().urn))

    exposed_thing.add_property(uuid.uuid4().hex, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().sentence())

    exposed_thing.add_event(uuid.uuid4().hex, EventFragmentDict({
        "type": "number"
    }))

    action_name = uuid.uuid4().hex

    @tornado.gen.coroutine
    def handler(parameters):
        input_value = parameters.get("input")
        yield tornado.gen.sleep(random.random() * 0.1)
        raise tornado.gen.Return("{:f}".format(input_value))

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "number"},
        "output": {"type": "string"}
    }), handler)

    server = MQTTServer(broker_url=get_test_broker_url(), **request.param)
    server.add_exposed_thing(exposed_thing)

    @tornado.gen.coroutine
    def start():
        yield server.start()

    tornado.ioloop.IOLoop.current().run_sync(start)

    yield server

    @tornado.gen.coroutine
    def stop():
        yield server.stop()

    tornado.ioloop.IOLoop.current().run_sync(stop)


@pytest.fixture
def mqtt_servient():
    """Returns a Servient that exposes a CoAP server and one ExposedThing."""

    from wotpy.protocols.mqtt.server import MQTTServer
    from tests.protocols.mqtt.broker import get_test_broker_url

    server = MQTTServer(broker_url=get_test_broker_url())

    servient = Servient(catalogue_port=None)
    servient.add_server(server)

    @tornado.gen.coroutine
    def start():
        raise tornado.gen.Return((yield servient.start()))

    wot = tornado.ioloop.IOLoop.current().run_sync(start)

    property_name_01 = uuid.uuid4().hex
    action_name_01 = uuid.uuid4().hex
    event_name_01 = uuid.uuid4().hex

    td_dict = {
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            property_name_01: {
                "observable": True,
                "type": "string"
            }
        },
        "actions": {
            action_name_01: {
                "input": {
                    "type": "number"
                },
                "output": {
                    "type": "number"
                },
            }
        },
        "events": {
            event_name_01: {
                "type": "string"
            }
        },
    }

    td = ThingDescription(td_dict)

    exposed_thing = wot.produce(td.to_str())
    exposed_thing.expose()

    @tornado.gen.coroutine
    def action_handler(parameters):
        input_value = parameters.get("input")
        raise tornado.gen.Return(int(input_value) * 2)

    exposed_thing.set_action_handler(action_name_01, action_handler)

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Request handler for Event interactions.
"""

import logging

import tornado.gen
from tornado.concurrent import Future
from tornado.web import RequestHandler

import wotpy.protocols.http.handlers.utils as handler_utils


# noinspection PyAbstractClass,PyAttributeOutsideInit
class EventObserverHandler(RequestHandler):
    """Handler for Event subscription requests."""

    # noinspection PyMethodOverriding
    def initialize(self, http_server):
        self._server = http_server
        self._logr = logging.getLogger(__name__)

    @tornado.gen.coroutine
    def get(self, thing_name, name):
        """Subscribes to the given Event and waits for the next emission (HTTP long-polling pattern).
        Returns the event emission payload and destroys the subscription afterwards."""

        exposed_thing = handler_utils.get_exposed_thing(self._server, thing_name)
        thing_event = exposed_thing.events[name]

        future_next = Future()

        def on_next(item):
            not future_next.done() and future_next.set_result(item.data)

        def on_error(err):
            self._logr.warning("Error on subscription to {}: {}".format(thing_event, err))
            not future_next.done() and future_next.set_exception(err)

        self.subscription = thing_event.subscribe(on_next=on_next, on_error=on_error)
        event_payload = yield future_next
        self.write({"payload": event_payload})

    def on_finish(self):
        """Destroys the subscription to the observable when the request finishes."""

        try:
            self.subscription.dispose()
        except AttributeError:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Enumeration classes related to the MQTT protocol binding.
"""

from wotpy.utils.enums import EnumListMixin


class MQTTSchemes(EnumListMixin):
    """Enumeration of MQTT schemes."""

    MQTT = "mqtt"


class MQTTCommandCodes(EnumListMixin):
    """Enumeration of MQTT packet types."""

    PUBLISH = 3
    SUBSCRIBE = 8
    UNSUBSCRIBE = 10


class MQTTQoSLevels(EnumListMixin):
    """Enumeration of MQTT Quality of Service levels."""

    FIRE_FORGET = 0
    AT_LEAST_ONCE = 1
    EXACTLY_ONCE = 2


class MQTTVocabularyKeys(EnumListMixin):
    """Enumeration of terms that form the MQTT vocabulary that may appear in TD Form elements."""

    COMMAND_CODE = "mqtt:commandCode"
    OPTIONS = "mqtt:options"
    OPTION_NAME = "mqtt:optionName"
    OPTION_VALUE = "mqtt:optionValue"
    OPTION_NAME_QOS = "mqtt:qos"
    OPTION_NAME_RETAIN = "mqtt:retain"
    OPTION_NAME_DUP = "mqtt:dup"


class MQTTCodesACK(EnumListMixin):
    """Enumeration of MQTT ACK codes."""

    CON_OK = 0
    SUB_ERROR = 128



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper classes for dictionaries for interaction initialization that are defined in the Scripting API.
"""

import six

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.wot.dictionaries.link import FormDict
from wotpy.wot.dictionaries.schema import DataSchemaDict
from wotpy.wot.dictionaries.security import SecuritySchemeDict


class InteractionFragmentDict(WotBaseDict):
    """Base class for the three types of Interaction patterns
    (Properties, Actions and Events)."""

    class Meta:
        fields = {
            "forms",
            "title",
            "uriVariables",
            "description",
            "security",
            "scopes"
        }

    @property
    def forms(self):
        """Indicates one or more endpoints from which
        an interaction pattern is accessible."""

        return [FormDict(item) for item in self._init.get("forms", [])]

    @property
    def uri_variables(self):
        """Define URI template variables as collection based on DataSchema declarations."""

        if "uriVariables" not in self._init:
            return None

        return {
            key: DataSchemaDict.build(val)
            for key, val in six.iteritems(self._init.get("uriVariables"))
        }

    @property
    def security(self):
        """Set of security configurations, provided as an array,
        that must all be satisfied for access to resources at or
        below the current level, if not overridden at a lower level."""

        if "security" not in self._init:
            return None

        return [SecuritySchemeDict.build(item) for item in self._init.get("security")]


class PropertyFragmentDict(InteractionFragmentDict):
    """A dictionary wrapper class that contains data to initialize a Property."""

    class Meta:
        fields = InteractionFragmentDict.Meta.fields.union({
            "observable"
        })

    def __init__(self, *args, **kwargs):
        super(PropertyFragmentDict, self).__init__(*args, **kwargs)
        self._data_schema = DataSchemaDict.build(self._init)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the internal ValueType before propagating the exception."""

        try:
            return super(PropertyFragmentDict, self).__getattr__(name)
        except AttributeError:
            return getattr(self.data_schema, name)

    def to_dict(self):
        """Returns the pure dict (JSON-serializable) representation of this WoT dictionary."""

        ret = super(PropertyFragmentDict, self).to_dict()
        ret.update(self.data_schema.to_dict())

        return ret

    @property
    def data_schema(self):
        """The DataSchema that represents the schema of this property."""

        return self._data_schema

    @property
    def writable(self):
        """Returns True if this Property is writable."""

        return not self.data_schema.read_only


class ActionFragmentDict(InteractionFragmentDict):
    """A dictionary wrapper class that contains data to initialize an Action."""

    class Meta:
        fields = InteractionFragmentDict.Meta.fields.union({
            "input",
            "output",
            "safe",
            "idempotent"
        })

        defaults = {
            "safe": False,
            "idempotent": False
        }

    @property
    def input(self):
        """Used to define the input data schema of the action."""

        init = self._init.get("input")

        return DataSchemaDict.build(init) if init else None

    @property
    def output(self):
        """Used to define the output data schema of the action."""

        init = self._init.get("output")

        return DataSchemaDict.build(init) if init else None


class EventFragmentDict(InteractionFragmentDict):
    """A dictionary wrapper class that contains data to initialize an Event."""

    class Meta:
        fields = InteractionFragmentDict.Meta.fields.union({
            "subscription",
            "data",
            "cancellation"
        })

    @property
    def subscription(self):
        """Defines data that needs to be passed upon subscription,
        e.g., filters or message format for setting up Webhooks."""

        init = self._init.get("subscription")

        return DataSchemaDict.build(init) if init else None

    @property
    def data(self):
        """Defines the data schema of the Event instance messages pushed by the Thing."""

        init = self._init.get("data")

        return DataSchemaDict.build(init) if init else None

    @property
    def cancellation(self):
        """Defines any data that needs to be passed to cancel a subscription,
        e.g., a specific message to remove a Webhook."""

        init = self._init.get("cancellation")

        return DataSchemaDict.build(init) if init else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that implements the MQTT server (broker).
"""

import tornado.gen
import tornado.ioloop
import tornado.locks
from slugify import slugify

from wotpy.codecs.enums import MediaTypes
from wotpy.protocols.enums import Protocols, InteractionVerbs
from wotpy.protocols.mqtt.handlers.action import ActionMQTTHandler
from wotpy.protocols.mqtt.handlers.event import EventMQTTHandler
from wotpy.protocols.mqtt.handlers.ping import PingMQTTHandler
from wotpy.protocols.mqtt.handlers.property import PropertyMQTTHandler
from wotpy.protocols.mqtt.runner import MQTTHandlerRunner
from wotpy.protocols.server import BaseProtocolServer
from wotpy.wot.enums import InteractionTypes
from wotpy.wot.form import Form


class MQTTServer(BaseProtocolServer):
    """MQTT binding server implementation."""

    DEFAULT_SERVIENT_ID = 'wotpy'

    def __init__(self, broker_url, property_callback_ms=None, event_callback_ms=None, servient_id=None):
        super(MQTTServer, self).__init__(port=None)
        self._broker_url = broker_url
        self._server_lock = tornado.locks.Lock()
        self._servient_id = servient_id

        def build_runner(handler):
            return MQTTHandlerRunner(broker_url=self._broker_url, mqtt_handler=handler)

        self._handler_runners = [
            build_runner(PingMQTTHandler(mqtt_server=self)),
            build_runner(PropertyMQTTHandler(mqtt_server=self, callback_ms=property_callback_ms)),
            build_runner(EventMQTTHandler(mqtt_server=self, callback_ms=event_callback_ms)),
            build_runner(ActionMQTTHandler(mqtt_server=self)),
        ]

    @property
    def servient_id(self):
        """Servient ID that is used to avoid topic collisions
        øwhen multiple Servients are connected to the same broker."""

        return slugify(self._servient_id) if self._servient_id else self.DEFAULT_SERVIENT_ID

    @property
    def protocol(self):
        """Protocol of this server instance.
        A member of the Protocols enum."""

        return Protocols.MQTT

    def _build_forms_property(self, proprty):
        """Builds and returns the MQTT Form instances for the given Property interaction."""

        href_rw = "{}/{}/property/requests/{}/{}".format(
            self._broker_url.rstrip("/"),
            self.servient_id,
            proprty.thing.url_name,
            proprty.url_name)

        form_read = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_rw,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.READ_PROPERTY)

        form_write = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_rw,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.WRITE_PROPERTY)

        href_observe = "{}/{}/property/updates/{}/{}".format(
            self._broker_url.rstrip("/"),
            self.servient_id,
            proprty.thing.url_name,
            proprty.url_name)

        form_observe = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_observe,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.OBSERVE_PROPERTY)

        return [form_read, form_write, form_observe]

    def _build_forms_action(self, action):
        """Builds and returns the MQTT Form instances for the given Action interaction."""

        href = "{}/{}/action/invocation/{}/{}".format(
            self._broker_url.rstrip("/"),
            self.servient_id,
            action.thing.url_name,
            action.url_name)

        form = Form(
            interaction=action,
            protocol=self.protocol,
            href=href,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.INVOKE_ACTION)

        return [form]

    def _build_forms_event(self, event):
        """Builds and returns the MQTT Form instances for the given Event interaction."""

        href = "{}/{}/event/{}/{}".format(
            self._broker_url.rstrip("/"),
            self.servient_id,
            event.thing.url_name,
            event.url_name)

        form = Form(
            interaction=event,
            protocol=self.protocol,
            href=href,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.SUBSCRIBE_EVENT)

        return [form]

    def build_forms(self, hostname, interaction):
        """Builds and returns a list with all Forms that are
        linked to this server for the given Interaction."""

        intrct_type_map = {
            InteractionTypes.PROPERTY: self._build_forms_property,
            InteractionTypes.ACTION: self._build_forms_action,
            InteractionTypes.EVENT: self._build_forms_event
        }

        if interaction.interaction_type not in intrct_type_map:
            raise ValueError("Unsupported interaction")

        return intrct_type_map[interaction.interaction_type](interaction)

    def build_base_url(self, hostname, thing):
        """Returns the base URL for the given Thing in the context of this server."""

        return self._broker_url

    @tornado.gen.coroutine
    def start(self):
        """Starts the MQTT broker and all the MQTT clients
        that handle the WoT clients requests."""

        with (yield self._server_lock.acquire()):
            yield [runner.start() for runner in self._handler_runners]

    @tornado.gen.coroutine
    def stop(self):
        """Stops the MQTT broker and the MQTT clients."""

        with (yield self._server_lock.acquire()):
            yield [runner.stop() for runner in self._handler_runners]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Exceptions raised by the protocol binding implementations.
"""


class ProtocolClientException(Exception):
    """Base Exceptions raised by clients of the protocol binding implementations."""

    DEFAULT_MSG = "Protocol client error"

    def __init__(self, *args, **kwargs):
        if not (args or kwargs):
            args = (self.DEFAULT_MSG,)

        super(ProtocolClientException, self).__init__(*args, **kwargs)


class FormNotFoundException(ProtocolClientException):
    """Exception raised when a form for a given protocol
    binding could not be found in a Thing Description."""

    DEFAULT_MSG = "Protocol Form not found in TD"


class ClientRequestTimeout(ProtocolClientException):
    """Exception raised when a protocol client request reaches the timeout."""

    DEFAULT_MSG = "Timeout in protocol client request"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents the codec interface.
"""


class BaseCodec(object):
    """Base codec abstract class.
    All codecs must implement this interface."""

    @property
    def media_types(self):
        """Property getter for the supported media types of this codec."""

        raise NotImplementedError()

    def to_value(self, value):
        """Takes an encoded value from a request that may be an UTF8
        bytes or unicode string and decodes it to a Python object."""

        raise NotImplementedError()

    def to_bytes(self, value):
        """Takes a Python object and encodes it to an UTF8
        bytes string to be included in a response."""

        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Enumeration classes related to the CoAP server.
"""

from wotpy.utils.enums import EnumListMixin


class CoAPSchemes(EnumListMixin):
    """Enumeration of CoAP schemes."""

    COAP = "coap"
    COAPS = "coaps"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that implements the HTTP server.
"""

import tornado.gen
import tornado.httpserver
import tornado.web

from wotpy.codecs.enums import MediaTypes
from wotpy.protocols.enums import Protocols, InteractionVerbs
from wotpy.protocols.http.enums import HTTPSchemes
from wotpy.protocols.http.handlers.action import ActionInvokeHandler, PendingInvocationHandler
from wotpy.protocols.http.handlers.event import EventObserverHandler
from wotpy.protocols.http.handlers.property import PropertyObserverHandler, PropertyReadWriteHandler
from wotpy.protocols.server import BaseProtocolServer
from wotpy.wot.enums import InteractionTypes
from wotpy.wot.form import Form


class HTTPServer(BaseProtocolServer):
    """HTTP binding server implementation."""

    DEFAULT_PORT = 80

    def __init__(self, port=DEFAULT_PORT, ssl_context=None, action_ttl_secs=300):
        super(HTTPServer, self).__init__(port=port)
        self._server = None
        self._app = self._build_app()
        self._ssl_context = ssl_context
        self._action_ttl_secs = action_ttl_secs
        self._pending_actions = {}
        self._invocation_check_times = {}

    @property
    def protocol(self):
        """Protocol of this server instance.
        A member of the Protocols enum."""

        return Protocols.HTTP

    @property
    def scheme(self):
        """Returns the URL scheme for this server."""

        return HTTPSchemes.HTTPS if self.is_secure else HTTPSchemes.HTTP

    @property
    def is_secure(self):
        """Returns True if this server is configured to use SSL encryption."""

        return self._ssl_context is not None

    @property
    def app(self):
        """Tornado application."""

        return self._app

    @property
    def action_ttl(self):
        """Returns the Action invocations Time-To-Live (seconds)."""

        return self._action_ttl_secs

    @property
    def pending_actions(self):
        """Dict of pending action invocations represented as Futures."""

        return self._pending_actions

    @property
    def invocation_check_times(self):
        """Dict that contains the timestamp of the last time an invocation was checked by a client.."""

        return self._invocation_check_times

    def _build_app(self):
        """Builds and returns the Tornado application for the WebSockets server."""

        return tornado.web.Application([(
            r"/(?P<thing_name>[^\/]+)/property/(?P<name>[^\/]+)",
            PropertyReadWriteHandler,
            {"http_server": self}
        ), (
            r"/(?P<thing_name>[^\/]+)/property/(?P<name>[^\/]+)/subscription",
            PropertyObserverHandler,
            {"http_server": self}
        ), (
            r"/(?P<thing_name>[^\/]+)/action/(?P<name>[^\/]+)",
            ActionInvokeHandler,
            {"http_server": self}
        ), (
            r"/invocation/(?P<invocation_id>[^\/]+)",
            PendingInvocationHandler,
            {"http_server": self}
        ), (
            r"/(?P<thing_name>[^\/]+)/event/(?P<name>[^\/]+)/subscription",
            EventObserverHandler,
            {"http_server": self}
        )])

    def _build_forms_property(self, proprty, hostname):
        """Builds and returns the HTTP Form instances for the given Property interaction."""

        href_read_write = "{}://{}:{}/{}/property/{}".format(
            self.scheme, hostname.rstrip("/").lstrip("/"), self.port,
            proprty.thing.url_name, proprty.url_name)

        form_read = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_read_write,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.READ_PROPERTY)

        form_write = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_read_write,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.WRITE_PROPERTY)

        href_observe = "{}/subscription".format(href_read_write)

        form_observe = Form(
            interaction=proprty,
            protocol=self.protocol,
            href=href_observe,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.OBSERVE_PROPERTY)

        return [form_read, form_write, form_observe]

    def _build_forms_action(self, action, hostname):
        """Builds and returns the HTTP Form instances for the given Action interaction."""

        href_invoke = "{}://{}:{}/{}/action/{}".format(
            self.scheme, hostname.rstrip("/").lstrip("/"), self.port,
            action.thing.url_name, action.url_name)

        form_invoke = Form(
            interaction=action,
            protocol=self.protocol,
            href=href_invoke,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.INVOKE_ACTION)

        return [form_invoke]

    def _build_forms_event(self, event, hostname):
        """Builds and returns the HTTP Form instances for the given Event interaction."""

        href_observe = "{}://{}:{}/{}/event/{}/subscription".format(
            self.scheme, hostname.rstrip("/").lstrip("/"), self.port,
            event.thing.url_name, event.url_name)

        form_observe = Form(
            interaction=event,
            protocol=self.protocol,
            href=href_observe,
            content_type=MediaTypes.JSON,
            op=InteractionVerbs.SUBSCRIBE_EVENT)

        return [form_observe]

    def build_forms(self, hostname, interaction):
        """Builds and returns a list with all Form that are
        linked to this server for the given Interaction."""

        intrct_type_map = {
            InteractionTypes.PROPERTY: self._build_forms_property,
            InteractionTypes.ACTION: self._build_forms_action,
            InteractionTypes.EVENT: self._build_forms_event
        }

        if interaction.interaction_type not in intrct_type_map:
            raise ValueError("Unsupported interaction")

        return intrct_type_map[interaction.interaction_type](interaction, hostname)

    def build_base_url(self, hostname, thing):
        """Returns the base URL for the given Thing in the context of this server."""

        if not self.exposed_thing_set.find_by_thing_id(thing.id):
            raise ValueError("Unknown Thing")

        return "{}://{}:{}/{}".format(
            self.scheme, hostname.rstrip("/").lstrip("/"),
            self.port, thing.url_name)

    @tornado.gen.coroutine
    def start(self):
        """Starts the HTTP server."""

        self._server = tornado.httpserver.HTTPServer(self.app, ssl_options=self._ssl_context)
        self._server.listen(self.port)

    @tornado.gen.coroutine
    def stop(self):
        """Stops the HTTP server."""

        if not self._server:
            return

        self._server.stop()
        self._server = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that contain the client logic for the MQTT protocol.
"""

import asyncio
import copy
import datetime
import json
import logging
import pprint
import time
import uuid

import hbmqtt.client
import tornado.concurrent
import tornado.gen
import tornado.ioloop
import tornado.locks
from hbmqtt.mqtt.constants import QOS_1, QOS_0, QOS_2
from rx import Observable
from six.moves.urllib import parse

from wotpy.protocols.client import BaseProtocolClient
from wotpy.protocols.enums import Protocols, InteractionVerbs
from wotpy.protocols.exceptions import FormNotFoundException, ClientRequestTimeout
from wotpy.protocols.mqtt.enums import MQTTSchemes
from wotpy.protocols.mqtt.handlers.action import ActionMQTTHandler
from wotpy.protocols.mqtt.handlers.property import PropertyMQTTHandler
from wotpy.protocols.refs import ConnRefCounter
from wotpy.protocols.utils import is_scheme_form
from wotpy.utils.utils import handle_observer_finalization
from wotpy.wot.events import PropertyChangeEventInit, PropertyChangeEmittedEvent, EmittedEvent


class MQTTClient(BaseProtocolClient):
    """Implementation of the protocol client interface for the MQTT protocol."""

    DELIVER_TERMINATE_LOOP_SLEEP_SECS = 0.1
    SLEEP_SECS_DELIVER_ERR = 1.0

    DEFAULT_DELIVER_TIMEOUT_SECS = 1
    DEFAULT_MSG_WAIT_TIMEOUT_SECS = 5
    DEFAULT_MSG_TTL_SECS = 15
    DEFAULT_STOP_LOOP_TIMEOUT_SECS = 60

    # Highly permissive default keep_alive to avoid
    # disconnections from broker on high throughput scenarios:
    # https://github.com/beerfactory/hbmqtt/issues/119#issuecomment-430398094

    DEFAULT_CLIENT_CONFIG = {
        "keep_alive": 90
    }

    def __init__(self,
                 deliver_timeout_secs=DEFAULT_DELIVER_TIMEOUT_SECS,
                 msg_wait_timeout_secs=DEFAULT_MSG_WAIT_TIMEOUT_SECS,
                 msg_ttl_secs=DEFAULT_MSG_TTL_SECS,
                 timeout_default=None,
                 hbmqtt_config=None,
                 stop_loop_timeout_secs=DEFAULT_STOP_LOOP_TIMEOUT_SECS):
        self._deliver_timeout_secs = deliver_timeout_secs
        self._msg_wait_timeout_secs = msg_wait_timeout_secs
        self._msg_ttl_secs = msg_ttl_secs
        self._timeout_default = timeout_default
        self._hbmqtt_config = hbmqtt_config
        self._stop_loop_timeout_secs = stop_loop_timeout_secs
        self._lock_client = tornado.locks.Lock()
        self._deliver_stop_events = {}
        self._msg_conditions = {}
        self._clients = {}
        self._messages = {}
        self._topics = {}
        self._ref_counter = ConnRefCounter()
        self._logr = logging.getLogger(__name__)

    def _build_client_config(self):
        """Returns the config dict for a new hbmqtt client instance."""

        config = copy.copy(self.DEFAULT_CLIENT_CONFIG)
        config_arg = self._hbmqtt_config if self._hbmqtt_config else {}
        config.update(config_arg)

        # The library does not resubscribe when reconnecting.
        # We need to handle it manually.

        config.update({"auto_reconnect": False})

        return config

    def _new_message(self, broker_url, msg):
        """Adds the message to the internal queue and notifies all topic listeners."""

        assert broker_url in self._msg_conditions, "Unknown broker in conditions"
        assert msg.topic in self._msg_conditions[broker_url], "Unknown topic"

        if broker_url not in self._messages:
            self._messages[broker_url] = {}

        if msg.topic not in self._messages[broker_url]:
            self._messages[broker_url][msg.topic] = []

        self._messages[broker_url][msg.topic].append({
            "id": uuid.uuid4().hex,
            "data": json.loads(msg.data.decode()),
            "time": time.time()
        })

        self._msg_conditions[broker_url][msg.topic].notify_all()
        self._clean_messages(broker_url)

    @tornado.gen.coroutine
    def _reconnect_client(self, broker_url):
        """Reconnects an existing client that has been disconnected."""

        assert broker_url in self._clients, "Unknown broker"

        self._logr.info("Reconnecting MQTT client: {}".format(broker_url))

        yield self._clients[broker_url].reconnect(cleansession=False)

        topics = self._topics.get(broker_url, set())

        if not len(topics):
            return

        self._logr.info("Resubscribing MQTT client on {} to topics:\n{}".format(
            broker_url, pprint.pformat(topics)))

        yield self._clients[broker_url].subscribe([(topic, qos) for topic, qos in topics])

    def _build_deliver(self, broker_url, stop_event):
        """Factory for functions to get messages delivered by the broker into the messages queue."""

        @tornado.gen.coroutine
        def reconnect():
            """Sleeps for a while and tries to reconnect and resubscribe afterwards."""

            try:
                self._logr.debug("Sleeping for {} s".format(self.SLEEP_SECS_DELIVER_ERR))
                yield tornado.gen.sleep(self.SLEEP_SECS_DELIVER_ERR)
                yield self._reconnect_client(broker_url)
            except Exception as ex_reconn:
                self._logr.warning("Error reconnecting: {}".format(ex_reconn), exc_info=True)

        @tornado.gen.coroutine
        def deliver():
            """Loop that receives the messages from the broker."""

            assert broker_url in self._clients

            self._logr.debug("Entering message delivery loop: {}".format(broker_url))

            while not stop_event.is_set():
                try:
                    msg = yield self._clients[broker_url].deliver_message(timeout=self._deliver_timeout_secs)
                except asyncio.TimeoutError:
                    continue
                except Exception as ex:
                    self._logr.warning("Error delivering message: {}".format(ex))
                    yield reconnect()
                    continue

                try:
                    self._new_message(broker_url, msg)
                except Exception as ex:
                    self._logr.warning("Error processing message: {}".format(ex), exc_info=True)

            self._logr.debug("Exiting message delivery loop: {}".format(broker_url))

            stop_event.clear()

        return deliver

    @tornado.gen.coroutine
    def _start_deliver_loop(self, broker_url):
        """Starts the message delivery loop in the background."""

        assert broker_url not in self._deliver_stop_events, "Stop event is already defined"

        stop_event = tornado.locks.Event()
        self._deliver_stop_events[broker_url] = stop_event
        deliver_loop_cb = self._build_deliver(broker_url, stop_event)
        tornado.ioloop.IOLoop.current().add_callback(deliver_loop_cb)

    @tornado.gen.coroutine
    def _stop_deliver_loop(self, broker_url):
        """Asks the message delivery loop to stop gracefully."""

        assert broker_url in self._deliver_stop_events, "Unknown broker"
        assert not self._deliver_stop_events[broker_url].is_set(), "Stop event is already set"

        self._deliver_stop_events[broker_url].set()

        now = time.time()

        def raise_timeout():
            """Checks if enought time has passed to raise a timeout error."""

            if self._stop_loop_timeout_secs is None:
                return

            if (time.time() - now) > self._stop_loop_timeout_secs:
                raise asyncio.TimeoutError("Timeout waiting for message delivery loop")

        while self._deliver_stop_events[broker_url].is_set():
            raise_timeout()
            yield tornado.gen.sleep(self.DELIVER_TERMINATE_LOOP_SLEEP_SECS)

        self._deliver_stop_events.pop(broker_url)

    @tornado.gen.coroutine
    def _init_client(self, broker_url, ref_id):
        """Initializes and connects a client to the given broker URL."""

        with (yield self._lock_client.acquire()):
            self._ref_counter.increase(broker_url, ref_id)

            if broker_url in self._clients:
                return

            config = self._build_client_config()

            self._logr.debug("Connecting MQTT client to {} with config: {}".format(
                broker_url, pprint.pformat(config)))

            self._clients[broker_url] = hbmqtt.client.MQTTClient(config=config)

            yield self._clients[broker_url].connect(broker_url, cleansession=False)

            yield self._start_deliver_loop(broker_url)

    @tornado.gen.coroutine
    def _disconnect_client(self, broker_url, ref_id):
        """Decreases the reference counter for the client on the given broker and cleans
        all resources when the client does not have any more references pointing to it."""

        with (yield self._lock_client.acquire()):
            self._ref_counter.decrease(broker_url, ref_id)

            if self._ref_counter.has_any(broker_url):
                return

            try:
                self._logr.debug("Stopping message delivery loop: {}".format(broker_url))
                yield self._stop_deliver_loop(broker_url)
            except Exception as ex:
                self._logr.warning("Error stopping deliver loop: {}".format(ex), exc_info=True)

            try:
                self._logr.debug("Disconnecting MQTT client: {}".format(broker_url))
                yield self._clients[broker_url].disconnect()
            except Exception as ex:
                self._logr.warning("Error disconnecting: {}".format(ex), exc_info=True)

            self._clients.pop(broker_url, None)
            self._messages.pop(broker_url, None)
            self._msg_conditions.pop(broker_url, None)
            self._topics.pop(broker_url, None)

    @tornado.gen.coroutine
    def _subscribe(self, broker_url, topic, qos):
        """Subscribes to a topic."""

        with (yield self._lock_client.acquire()):
            if broker_url not in self._clients:
                return

            if broker_url not in self._msg_conditions:
                self._msg_conditions[broker_url] = {}

            if topic not in self._msg_conditions[broker_url]:
                self._msg_conditions[broker_url][topic] = tornado.locks.Condition()

            if broker_url not in self._topics:
                self._topics[broker_url] = set()

            self._topics[broker_url].add((topic, qos))

            yield self._clients[broker_url].subscribe([(topic, qos)])

    @tornado.gen.coroutine
    def _publish(self, broker_url, topic, payload, qos):
        """Publishes a message with the given payload in a topic."""

        with (yield self._lock_client.acquire()):
            if broker_url not in self._clients:
                return

            yield self._clients[broker_url].publish(topic, payload, qos=qos)

    def _topic_messages(self, broker_url, topic, from_time=None, ignore_ids=None):
        """Returns a generator that yields the messages in the
        delivered messages queue for the given topic."""

        if broker_url not in self._messages:
            return

        if topic not in self._messages[broker_url]:
            return

        for msg in self._messages[broker_url][topic]:
            is_on_time = from_time is None or msg["time"] >= from_time
            is_ignored = ignore_ids is not None and msg["id"] in ignore_ids

            if is_on_time and not is_ignored:
                yield msg["id"], msg["data"]

    def _clean_messages(self, broker_url):
        """Removes the messages that have expired according to the TTL."""

        if broker_url not in self._messages:
            return

        now = time.time()

        self._messages[broker_url] = {
            topic: [
                msg for msg in self._messages[broker_url][topic]
                if (now - msg["time"]) < self._msg_ttl_secs
            ] for topic in self._messages[broker_url]
        }

    def _next_match(self, broker_url, topic, func):
        """Returns the first message match in the internal messages queue or None."""

        return next((item for item in self._topic_messages(broker_url, topic) if func(item)), None)

    @tornado.gen.coroutine
    def _wait_on_message(self, broker_url, topic):
        """Waits for the arrival of a message in the given topic."""

        assert broker_url in self._msg_conditions, "Unknown broker URL"
        assert topic in self._msg_conditions[broker_url], "Unknown topic"

        wait_timeout = datetime.timedelta(seconds=self._msg_wait_timeout_secs)

        yield self._msg_conditions[broker_url][topic].wait(timeout=wait_timeout)

    @classmethod
    def _pick_mqtt_href(cls, td, forms, op=None):
        """Picks the most appropriate MQTT form href from the given list of forms."""

        def is_op_form(form):
            try:
                return op is None or op == form.op or op in form.op
            except TypeError:
                return False

        return next((
            form.href for form in forms
            if is_scheme_form(form, td.base, MQTTSchemes.MQTT) and is_op_form(form)
        ), None)

    @classmethod
    def _parse_href(cls, href):
        """Takes an MQTT form href and returns
        the MQTT broker URL and the topic separately."""

        parsed_href = parse.urlparse(href)
        assert parsed_href.scheme and parsed_href.netloc and parsed_href.path

        return {
            "broker_url": "{}://{}".format(parsed_href.scheme, parsed_href.netloc),
            "topic": parsed_href.path.lstrip("/").rstrip("/")
        }

    @property
    def protocol(self):
        """Protocol of this client instance.
        A member of the Protocols enum."""

        return Protocols.MQTT

    def is_supported_interaction(self, td, name):
        """Returns True if the any of the Forms for the Interaction
        with the given name is supported in this Protocol Binding client."""

        forms = td.get_forms(name)

        forms_mqtt = [
            form for form in forms
            if is_scheme_form(form, td.base, MQTTSchemes.list())
        ]

        return len(forms_mqtt) > 0

    @tornado.gen.coroutine
    def invoke_action(self, td, name, input_value, timeout=None,
                      qos_publish=QOS_2, qos_subscribe=QOS_1):
        """Invokes an Action on a remote Thing.
        Returns a Future."""

        timeout = timeout if timeout else self._timeout_default
        ref_id = uuid.uuid4().hex

        href = self._pick_mqtt_href(td, td.get_action_forms(name))

        if href is None:
            raise FormNotFoundException()

        parsed_href = self._parse_href(href)
        broker_url = parsed_href["broker_url"]

        topic_invoke = parsed_href["topic"]
        topic_result = ActionMQTTHandler.to_result_topic(topic_invoke)

        try:
            yield self._init_client(broker_url, ref_id)
            yield self._subscribe(broker_url, topic_result, qos_subscribe)

            input_data = {
                "id": uuid.uuid4().hex,
                "input": input_value
            }

            input_payload = json.dumps(input_data).encode()

            yield self._publish(broker_url, topic_invoke, input_payload, qos_publish)

            ini = time.time()

            while True:
                self._logr.debug("Checking invocation topic: {}".format(topic_result))

                if timeout and (time.time() - ini) > timeout:
                    self._logr.warning("Timeout invoking Action: {}".format(topic_result))
                    raise ClientRequestTimeout

                msg_match = self._next_match(
                    broker_url, topic_result,
                    lambda item: item[1].get("id") == input_data.get("id"))

                if not msg_match:
                    yield self._wait_on_message(broker_url, topic_result)
                    continue

                msg_id, msg_data = msg_match

                if msg_data.get("error", None) is not None:
                    raise Exception(msg_data.get("error"))
                else:
                    raise tornado.gen.Return(msg_data.get("result"))
        finally:
            yield self._disconnect_client(broker_url, ref_id)

    @tornado.gen.coroutine
    def write_property(self, td, name, value, timeout=None,
                       qos_publish=QOS_2, qos_subscribe=QOS_1, wait_ack=True):
        """Updates the value of a Property on a remote Thing.
        Due to the MQTT binding design this coroutine yields as soon as the write message has
        been published and will not wait for a custom write handler that yields to another coroutine
        Returns a Future."""

        timeout = timeout if timeout else self._timeout_default
        ref_id = uuid.uuid4().hex

        href_write = self._pick_mqtt_href(
            td, td.get_property_forms(name),
            op=InteractionVerbs.WRITE_PROPERTY)

        if href_write is None:
            raise FormNotFoundException()

        parsed_href_write = self._parse_href(href_write)
        broker_url = parsed_href_write["broker_url"]

        topic_write = parsed_href_write["topic"]
        topic_ack = PropertyMQTTHandler.to_write_ack_topic(topic_write)

        try:
            yield self._init_client(broker_url, ref_id)
            yield self._subscribe(broker_url, topic_ack, qos_subscribe)

            write_data = {
                "action": "write",
                "value": value,
                "ack": uuid.uuid4().hex
            }

            write_payload = json.dumps(write_data).encode()

            yield self._publish(broker_url, topic_write, write_payload, qos_publish)

            if not wait_ack:
                return

            ini = time.time()

            while True:
                self._logr.debug("Checking write ACK topic: {}".format(topic_ack))

                if timeout and (time.time() - ini) > timeout:
                    self._logr.warning("Timeout writing Property: {}".format(topic_ack))
                    raise ClientRequestTimeout

                msg_match = self._next_match(
                    broker_url, topic_ack,
                    lambda item: item[1].get("ack") == write_data.get("ack"))

                if msg_match:
                    break

                yield self._wait_on_message(broker_url, topic_ack)
        finally:
            yield self._disconnect_client(broker_url, ref_id)

    @tornado.gen.coroutine
    def read_property(self, td, name, timeout=None,
                      qos_publish=QOS_1, qos_subscribe=QOS_1):
        """Reads the value of a Property on a remote Thing.
        Returns a Future."""

        timeout = timeout if timeout else self._timeout_default
        ref_id = uuid.uuid4().hex

        forms = td.get_property_forms(name)

        href_read = self._pick_mqtt_href(td, forms, op=InteractionVerbs.READ_PROPERTY)
        href_obsv = self._pick_mqtt_href(td, forms, op=InteractionVerbs.OBSERVE_PROPERTY)

        if href_read is None or href_obsv is None:
            raise FormNotFoundException()

        parsed_href_read = self._parse_href(href_read)
        parsed_href_obsv = self._parse_href(href_obsv)

        topic_read = parsed_href_read["topic"]
        topic_obsv = parsed_href_obsv["topic"]

        broker_read = parsed_href_read["broker_url"]
        broker_obsv = parsed_href_obsv["broker_url"]

        try:
            yield self._init_client(broker_read, ref_id)
            broker_obsv != broker_read and (yield self._init_client(broker_obsv, ref_id))

            yield self._subscribe(broker_obsv, topic_obsv, qos_subscribe)

            read_time = int(time.time() * 1000)
            read_payload = json.dumps({"action": "read"}).encode()

            yield self._publish(broker_read, topic_read, read_payload, qos_publish)

            ini = time.time()

            while True:
                self._logr.debug("Checking property update topic: {}".format(topic_obsv))

                if timeout and (time.time() - ini) > timeout:
                    self._logr.warning("Timeout reading Property: {}".format(topic_obsv))
                    raise ClientRequestTimeout

                msg_match = self._next_match(
                    broker_obsv, topic_obsv,
                    lambda item: item[1].get("timestamp") >= read_time)

                if not msg_match:
                    yield self._wait_on_message(broker_obsv, topic_obsv)
                    continue

                msg_id, msg_data = msg_match

                raise tornado.gen.Return(msg_data.get("value"))
        finally:
            yield self._disconnect_client(broker_read, ref_id)
            broker_obsv != broker_read and (yield self._disconnect_client(broker_obsv, ref_id))

    def _build_subscribe(self, broker_url, topic, next_item_builder, qos):
        """Builds the subscribe function that should be passed when
        constructing an Observable to listen for messages on an MQTT topic."""

        def subscribe(observer):
            """Subscriber function that listens for MQTT messages
            on a given topic and passes them to the Observer."""

            state = {"active": True}

            config = self._build_client_config()
            client = hbmqtt.client.MQTTClient(config=config)

            @handle_observer_finalization(observer)
            @tornado.gen.coroutine
            def callback():
                self._logr.debug("Subscribing on <{}> to {} with config: {}".format(
                    broker_url, topic, config))

                yield client.connect(broker_url)
                yield client.subscribe([(topic, qos)])

                while state["active"]:
                    try:
                        msg = yield client.deliver_message(timeout=self._deliver_timeout_secs)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        msg_data = json.loads(msg.data.decode())
                        next_item = next_item_builder(msg_data)
                        observer.on_next(next_item)
                    except Exception as ex:
                        self._logr.warning("Subscription message error: {}".format(ex), exc_info=True)

            def unsubscribe():
                """Disconnects from the MQTT broker and stops the message delivering loop."""

                @tornado.gen.coroutine
                def disconnect():
                    try:
                        yield client.disconnect()
                    except Exception as ex:
                        self._logr.warning("Subscription disconnection error: {}".format(ex))

                tornado.ioloop.IOLoop.current().add_callback(disconnect)

                state["active"] = False

            tornado.ioloop.IOLoop.current().add_callback(callback)

            return unsubscribe

        return subscribe

    def on_property_change(self, td, name, qos=QOS_0):
        """Subscribes to property changes on a remote Thing.
        Returns an Observable"""

        forms = td.get_property_forms(name)
        href = self._pick_mqtt_href(td, forms, op=InteractionVerbs.OBSERVE_PROPERTY)

        if href is None:
            raise FormNotFoundException()

        parsed_href = self._parse_href(href)

        broker_url = parsed_href["broker_url"]
        topic = parsed_href["topic"]

        def next_item_builder(msg_data):
            msg_value = msg_data.get("value")
            init = PropertyChangeEventInit(name=name, value=msg_value)
            return PropertyChangeEmittedEvent(init=init)

        subscribe = self._build_subscribe(
            broker_url=broker_url,
            topic=topic,
            next_item_builder=next_item_builder,
            qos=qos)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_event(self, td, name, qos=QOS_0):
        """Subscribes to an event on a remote Thing.
        Returns an Observable."""

        forms = td.get_event_forms(name)
        href = self._pick_mqtt_href(td, forms, op=InteractionVerbs.SUBSCRIBE_EVENT)

        if href is None:
            raise FormNotFoundException()

        parsed_href = self._parse_href(href)

        broker_url = parsed_href["broker_url"]
        topic = parsed_href["topic"]

        def next_item_builder(msg_data):
            return EmittedEvent(init=msg_data.get("data"), name=name)

        subscribe = self._build_subscribe(
            broker_url=broker_url,
            topic=topic,
            next_item_builder=next_item_builder,
            qos=qos)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_td_change(self, url):
        """Subscribes to Thing Description changes on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that represent all interaction patterns.
"""

from abc import ABCMeta, abstractmethod

# noinspection PyPackageRequirements
from slugify import slugify

from wotpy.wot.enums import InteractionTypes
from wotpy.wot.validation import is_valid_safe_name
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict, EventFragmentDict


class InteractionPattern(object):
    """A functionality exposed by Thing that is defined by the TD Interaction Model."""

    __metaclass__ = ABCMeta

    def __init__(self, thing, name, init_dict=None, **kwargs):
        if not is_valid_safe_name(name):
            raise ValueError("Invalid Interaction name: {}".format(name))

        self._init_dict = init_dict if init_dict else self.init_class(**kwargs)

        self._thing = thing
        self._name = name
        self._forms = []

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._init_dict, name)

    @property
    @abstractmethod
    def init_class(self):
        """Returns the init dict class for this type of interaction."""

        raise NotImplementedError()

    @property
    def interaction_fragment(self):
        """The InteractionFragment dictionary of this interaction."""

        return self._init_dict

    @property
    def thing(self):
        """Thing that contains this Interaction."""

        return self._thing

    @property
    def name(self):
        """Interaction name.
        No two Interactions with the same name may exist in a Thing."""

        return self._name

    @property
    def url_name(self):
        """URL-safe version of the name."""

        return slugify(self.name)

    @property
    def forms(self):
        """Sequence of forms linked to this interaction."""

        return self._forms

    def clean_forms(self):
        """Removes all the Forms from this Interaction."""

        self._forms = []

    def add_form(self, form):
        """Add a new Form."""

        assert form.interaction is self

        existing = next((True for item in self._forms if item.id == form.id), False)

        if existing:
            raise ValueError("Duplicate Form: {}".format(form))

        self._forms.append(form)

    def remove_form(self, form):
        """Remove an existing Form."""

        try:
            pop_idx = self._forms.index(form)
            self._forms.pop(pop_idx)
        except ValueError:
            pass


class Property(InteractionPattern):
    """Properties expose internal state of a Thing that can be
    directly accessed (get) and optionally manipulated (set)."""

    @property
    def init_class(self):
        """Returns the init dict class for this type of interaction."""

        return PropertyFragmentDict

    @property
    def interaction_type(self):
        """Interaction type."""

        return InteractionTypes.PROPERTY


class Action(InteractionPattern):
    """Actions offer functions of the Thing. These functions may manipulate the
    internal state of a Thing in a way that is not possible through setting Properties."""

    @property
    def init_class(self):
        """Returns the init dict class for this type of interaction."""

        return ActionFragmentDict

    @property
    def interaction_type(self):
        """Interaction type."""

        return InteractionTypes.ACTION


class Event(InteractionPattern):
    """The Event Interaction Pattern describes event sources that asynchronously push messages.
    Here not state, but state transitions (events) are communicated (e.g., clicked)."""

    @property
    def init_class(self):
        """Returns the init dict class for this type of interaction."""

        return EventFragmentDict

    @property
    def interaction_type(self):
        """Interaction type."""

        return InteractionTypes.EVENT



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Schemas following the JSON Schema specification used to validate the shape of Thing Description documents.
"""

import re

from wotpy.wot.enums import InteractionTypes

REGEX_SAFE_NAME = r"^[a-zA-Z0-9_-]+$"
REGEX_ANY_URI = r"^((\w+:(\/?\/?)[^\s]+)|((..\/)+)[^\s]*)$"

DATA_TYPES_ENUM = [
    "array",
    "boolean",
    "number",
    "integer",
    "object",
    "string",
    "null"
]

SCHEMA_DATA_SCHEMA = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/data-schema.json",
    "type": "object",
    "properties": {
        "description": {"type": "string"},
        "title": {"type": "string"},
        "type": {
            "type": "string",
            "enum": DATA_TYPES_ENUM
        },
        "const": {},
        "unit": {"type": "string"},
        "enum": {
            "type": "array",
            "items": {}
        },
        "readOnly": {"type": "boolean"},
        "writeOnly": {"type": "boolean"}
    },
    "required": [
        "type"
    ]
}

SCHEMA_SECURITY_SCHEME = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/security-scheme.json",
    "type": "object",
    "properties": {
        "scheme": {"type": "string"},
        "description": {"type": "string"},
        "proxyUrl": {"type": "string"}
    },
    "required": [
        "scheme"
    ]
}

SCHEMA_LINK = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/link.json",
    "type": "object",
    "properties": {
        "href": {
            "type": "string",
            "pattern": REGEX_ANY_URI
        },
        "type": {"type": "string"},
        "rel": {"type": "string"},
        "anchor": {
            "type": "string",
            "pattern": REGEX_ANY_URI
        },
    },
    "required": [
        "href"
    ]
}

SCHEMA_FORM = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/form.json",
    "type": "object",
    "properties": {
        "href": {"type": "string"},
        "contentType": {
            "type": "string",
            "default": "application/json"
        },
        "op": {"type": "string"},
        "subprotocol": {"type": "string"},
        "security": {
            "type": "array",
            "items": SCHEMA_SECURITY_SCHEME
        },
        "scopes": {
            "type": "array",
            "items": {"type": "string"}
        }
    },
    "required": [
        "href"
    ]
}

SCHEMA_INTERACTION_PATTERN = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/interaction-pattern.json",
    "type": "object",
    "properties": {
        "forms": {
            "type": "array",
            "items": SCHEMA_FORM
        },
        "title": {"type": "string"},
        "uriVariables": {
            "type": "object",
            "patternProperties": {REGEX_SAFE_NAME: SCHEMA_DATA_SCHEMA},
            "additionalProperties": False
        },
        "description": {"type": "string"},
        "security": {
            "type": "array",
            "items": SCHEMA_SECURITY_SCHEME
        },
        "scopes": {
            "type": "array",
            "items": {"type": "string"}
        }
    }
}

SCHEMA_PROPERTY = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/property.json",
    "allOf": [
        SCHEMA_INTERACTION_PATTERN,
        SCHEMA_DATA_SCHEMA,
        {
            "type": "object",
            "properties": {
                "observable": {
                    "type": "boolean",
                    "default": False
                }
            }
        }
    ]
}

SCHEMA_EVENT = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/event.json",
    "allOf": [
        SCHEMA_INTERACTION_PATTERN,
        {
            "type": "object",
            "properties": {
                "subscription": SCHEMA_DATA_SCHEMA,
                "data": SCHEMA_DATA_SCHEMA,
                "cancellation": SCHEMA_DATA_SCHEMA
            }
        }
    ]
}

SCHEMA_ACTION = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/action.json",
    "allOf": [
        SCHEMA_INTERACTION_PATTERN,
        {
            "type": "object",
            "properties": {
                "input": SCHEMA_DATA_SCHEMA,
                "output": SCHEMA_DATA_SCHEMA,
                "safe": {
                    "type": "boolean",
                    "default": False
                },
                "idempotent": {
                    "type": "boolean",
                    "default": False
                }
            }
        }
    ]
}

SCHEMA_VERSIONING = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/versioning.json",
    "type": "object",
    "properties": {
        "instance": {"type": "string"}
    },
    "required": [
        "instance"
    ]
}

SCHEMA_THING = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/thing.json",
    "type": "object",
    "properties": {
        "id": {
            "type": "string",
            "pattern": REGEX_ANY_URI
        },
        "version": SCHEMA_VERSIONING,
        "name": {"type": "string"},
        "description": {"type": "string"},
        "support": {"type": "string"},
        "created": {"type": "string"},
        "lastModified": {"type": "string"},
        "base": {
            "type": "string",
            "pattern": REGEX_ANY_URI
        },
        "properties": {
            "type": "object",
            "patternProperties": {REGEX_SAFE_NAME: SCHEMA_PROPERTY},
            "additionalProperties": False
        },
        "actions": {
            "type": "object",
            "patternProperties": {REGEX_SAFE_NAME: SCHEMA_ACTION},
            "additionalProperties": False
        },
        "events": {
            "type": "object",
            "patternProperties": {REGEX_SAFE_NAME: SCHEMA_EVENT},
            "additionalProperties": False
        },
        "links": {
            "type": "array",
            "items": SCHEMA_LINK
        },
        "security": {
            "type": "array",
            "items": SCHEMA_SECURITY_SCHEME
        }
    },
    "required": [
        "id",
        "name",
        "security"
    ]
}


def interaction_schema_for_type(interaction_type):
    """Returns the JSON schema that describes an
    interaction for the given interaction type."""

    type_schema_dict = {
        InteractionTypes.PROPERTY: SCHEMA_PROPERTY,
        InteractionTypes.ACTION: SCHEMA_ACTION,
        InteractionTypes.EVENT: SCHEMA_EVENT
    }

    assert interaction_type in type_schema_dict

    return type_schema_dict[interaction_type]


def is_valid_uri(val):
    """Returns True if the given value is a valid URI."""

    return False if re.match(REGEX_ANY_URI, val) is None else True


def is_valid_safe_name(val):
    """Returns True if the given value is a safe machine-readable name."""

    return False if re.match(REGEX_SAFE_NAME, val) is None else True


class InvalidDescription(Exception):
    """Exception raised when a document for an object
    in the TD hierarchy has an invalid format."""

    pass



import logging
import os

logger_base = logging.getLogger(__name__)
logger_base.addHandler(logging.NullHandler())

if os.environ.get("WOTPY_ENABLE_UVLOOP", False):
    try:
        import asyncio
        import uvloop

        logger_base.warning("Installing uvloop (this is an experimental feature)")
        uvloop.install()
    except ImportError:
        logger_base.warning("Error installing uvloop (cannot import package)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that implements the JSON codec.
"""

import json

from wotpy.codecs.base import BaseCodec
from wotpy.codecs.enums import MediaTypes


class JsonCodec(BaseCodec):
    """JSON codec class."""

    @property
    def media_types(self):
        """Returns the JSON media types."""

        return [MediaTypes.JSON]

    def to_value(self, value):
        """Takes an encoded value from a request that may be an UTF8 bytes
        or unicode JSON string and deserializes it to a Python object."""

        return json.loads(value)

    def to_bytes(self, value):
        """Takes an object and serializes it to an UTF8 bytes JSON string."""

        json_str = json.dumps(value)

        return json_str if isinstance(json_str, bytes) else json_str.encode('utf8')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper classes for link dictionaries defined in the Scripting API.
"""

from six.moves import urllib

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.wot.dictionaries.security import SecuritySchemeDict


class LinkDict(WotBaseDict):
    """A Web link, as specified by IETF RFC 8288."""

    class Meta:
        fields = {
            "href",
            "type",
            "rel",
            "anchor"
        }

        required = {
            "href"
        }


class FormDict(LinkDict):
    """Communication metadata indicating where a service can be accessed
    by a client application. An interaction might have more than one form."""

    class Meta:
        fields = LinkDict.Meta.fields.union({
            "href",
            "contentType",
            "op",
            "subprotocol",
            "security",
            "scopes"
        })

        required = LinkDict.Meta.required.union({
            "href"
        })

        defaults = {
            "contentType": "application/json"
        }

    @property
    def security(self):
        """Set of security configurations, provided as an array,
        that must all be satisfied for access to resources at or
        below the current level, if not overridden at a lower level"""

        if "security" not in self._init:
            return None

        return [SecuritySchemeDict.build(item) for item in self._init.get("security")]

    def resolve_uri(self, base=None):
        """Resolves and returns the Link URI.
        When the href does not contain a full URL the base URI is joined with said href."""

        href_parsed = urllib.parse.urlparse(self.href)

        if base and not href_parsed.scheme:
            return urllib.parse.urljoin(base, self.href)

        if href_parsed.scheme:
            return self.href

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that represent Interaction instances accessed on a ExposedThing.
"""

import six
import tornado.gen
from rx.concurrency import IOLoopScheduler
from six.moves import UserDict
from slugify import slugify


class ExposedThingInteractionDict(UserDict):
    """A dictionary that provides lazy access to the objects that implement
    the Interaction interface for each interaction in a given ExposedThing."""

    def __init__(self, *args, **kwargs):
        self._exposed_thing = kwargs.pop("exposed_thing")
        UserDict.__init__(self, *args, **kwargs)

    def _find_normalized_name(self, name):
        """Takes a case-insensitive URL-safe interaction name and returns
        the actual name in the interaction dict."""

        return next((key for key in six.iterkeys(self.interaction_dict) if slugify(key) == slugify(name)), None)

    def __getitem__(self, name):
        """Lazily build and return an object that implements the Interaction interface."""

        name_normalized = self._find_normalized_name(name)

        if name_normalized is None:
            raise KeyError("Unknown interaction: {}".format(name))

        return self.thing_interaction_class(self._exposed_thing, name_normalized)

    def __len__(self):
        return len(self.interaction_dict)

    def __contains__(self, item):
        return self._find_normalized_name(item) is not None

    def __iter__(self):
        return six.iterkeys(self.interaction_dict)

    @property
    def interaction_dict(self):
        """Returns the InteractionPattern objects dict by name."""

        raise NotImplementedError()

    @property
    def thing_interaction_class(self):
        """Returns the class that implements the
        Interaction interface for this type of interaction."""

        raise NotImplementedError()


class ExposedThingPropertyDict(ExposedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingProperty interface for each property in a given ExposedThing."""

    @property
    def interaction_dict(self):
        return self._exposed_thing.thing.properties

    @property
    def thing_interaction_class(self):
        return ExposedThingProperty


class ExposedThingActionDict(ExposedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingAction interface for each action in a given ExposedThing."""

    @property
    def interaction_dict(self):
        return self._exposed_thing.thing.actions

    @property
    def thing_interaction_class(self):
        return ExposedThingAction


class ExposedThingEventDict(ExposedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingEvent interface for each event in a given ExposedThing."""

    @property
    def interaction_dict(self):
        return self._exposed_thing.thing.events

    @property
    def thing_interaction_class(self):
        return ExposedThingEvent


class ExposedThingProperty(object):
    """The ThingProperty interface implementation for ExposedThing objects."""

    def __init__(self, exposed_thing, name):
        self._exposed_thing = exposed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._exposed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._exposed_thing.thing.properties[self._name], name)

    @tornado.gen.coroutine
    def read(self):
        """The get() method will fetch the value of the Property.
        A coroutine that yields the value or raises an error."""

        value = yield self._exposed_thing.read_property(self._name)
        raise tornado.gen.Return(value)

    @tornado.gen.coroutine
    def write(self, value):
        """The set() method will attempt to set the value of the
        Property specified in the value argument whose type SHOULD
        match the one specified by the type property.
        A coroutine that yields on success or raises an error."""

        yield self._exposed_thing.write_property(self._name, value)

    def subscribe(self, *args, **kwargs):
        """Subscribe to an stream of events emitted when the property value changes."""

        observable = self._exposed_thing.on_property_change(self._name)
        return observable.subscribe_on(IOLoopScheduler()).subscribe(*args, **kwargs)


class ExposedThingAction(object):
    """The ThingAction interface implementation for ExposedThing objects."""

    def __init__(self, exposed_thing, name):
        self._exposed_thing = exposed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._exposed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._exposed_thing.thing.actions[self._name], name)

    @tornado.gen.coroutine
    def invoke(self, *args):
        """The run() method when invoked, starts the Action interaction
        with the input value provided by the inputValue argument."""

        input_value = args[0] if len(args) else None
        result = yield self._exposed_thing.invoke_action(self._name, input_value)
        raise tornado.gen.Return(result)


class ExposedThingEvent(object):
    """The ThingEvent interface implementation for ExposedThing objects."""

    def __init__(self, exposed_thing, name):
        self._exposed_thing = exposed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._exposed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._exposed_thing.thing.events[self._name], name)

    def subscribe(self, *args, **kwargs):
        """Subscribe to an stream of emissions of this event."""

        observable = self._exposed_thing.on_event(self._name)
        return observable.subscribe_on(IOLoopScheduler()).subscribe(*args, **kwargs)

    def emit(self, payload):
        """Emits an event that carries data specified by the payload argument."""

        return self._exposed_thing.emit_event(self._name, payload)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility functions for CoAP resources.
"""

import six

from six.moves.urllib import parse


def parse_request_opt_query(request):
    """Takes a CoAP Request and returns a dict containing
    the parsed URI query parameters."""

    parsed_dict = parse.parse_qs("&".join(request.opt.uri_query))
    return {key: val[0] for key, val in six.iteritems(parsed_dict) if len(val)}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
DNS Service Discovery (based on Multicast DNS) Thing discovery service.

.. autosummary::
    :toctree: _dnssd

    wotpy.wot.discovery.dnssd.service
"""

from wotpy.support import is_dnssd_supported

if is_dnssd_supported() is False:
    raise NotImplementedError("DNS-SD is not supported in this platform")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Dictionaries and classes defined under the W3C WoT Scripting API specification.

.. autosummary::
    :toctree: _wot

    wotpy.wot.consumed
    wotpy.wot.dictionaries
    wotpy.wot.discovery
    wotpy.wot.exposed
    wotpy.wot.constants
    wotpy.wot.enums
    wotpy.wot.events
    wotpy.wot.form
    wotpy.wot.interaction
    wotpy.wot.servient
    wotpy.wot.td
    wotpy.wot.thing
    wotpy.wot.validation
    wotpy.wot.wot
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that contain the client logic for the HTTP protocol.
"""

import json
import logging
import time

import tornado.concurrent
import tornado.gen
import tornado.httpclient
import tornado.ioloop
from rx import Observable
from six.moves.urllib import parse
from tornado.simple_httpclient import HTTPTimeoutError

from wotpy.protocols.client import BaseProtocolClient
from wotpy.protocols.enums import Protocols, InteractionVerbs
from wotpy.protocols.exceptions import FormNotFoundException, ClientRequestTimeout
from wotpy.protocols.http.enums import HTTPSchemes
from wotpy.protocols.utils import is_scheme_form
from wotpy.utils.utils import handle_observer_finalization
from wotpy.wot.events import EmittedEvent, PropertyChangeEmittedEvent, PropertyChangeEventInit


class HTTPClient(BaseProtocolClient):
    """Implementation of the protocol client interface for the HTTP protocol."""

    JSON_HEADERS = {"Content-Type": "application/json"}
    DEFAULT_CON_TIMEOUT = 60
    DEFAULT_REQ_TIMEOUT = 60

    def __init__(self, connect_timeout=DEFAULT_CON_TIMEOUT, request_timeout=DEFAULT_REQ_TIMEOUT):
        self._connect_timeout = connect_timeout
        self._request_timeout = request_timeout
        self._logr = logging.getLogger(__name__)
        super(HTTPClient, self).__init__()

    @classmethod
    def pick_http_href(cls, td, forms, op=None):
        """Picks the most appropriate HTTP form href from the given list of forms."""

        def is_op_form(form):
            try:
                return op is None or op == form.op or op in form.op
            except TypeError:
                return False

        def find_href(scheme):
            try:
                return next(
                    form.href for form in forms
                    if is_scheme_form(form, td.base, scheme) and is_op_form(form))
            except StopIteration:
                return None

        form_https = find_href(HTTPSchemes.HTTPS)

        return form_https if form_https is not None else find_href(HTTPSchemes.HTTP)

    @property
    def protocol(self):
        """Protocol of this client instance.
        A member of the Protocols enum."""

        return Protocols.HTTP

    @property
    def connect_timeout(self):
        """Returns the default connection timeout for all HTTP requests."""

        return self._connect_timeout

    @property
    def request_timeout(self):
        """Returns the default request timeout for all HTTP requests."""

        return self._request_timeout

    def is_supported_interaction(self, td, name):
        """Returns True if the any of the Forms for the Interaction
        with the given name is supported in this Protocol Binding client."""

        forms = td.get_forms(name)

        forms_http = [
            form for form in forms
            if is_scheme_form(form, td.base, HTTPSchemes.list())
        ]

        return len(forms_http) > 0

    @tornado.gen.coroutine
    def invoke_action(self, td, name, input_value, timeout=None):
        """Invokes an Action on a remote Thing.
        Returns a Future."""

        con_timeout = timeout if timeout else self._connect_timeout
        req_timeout = timeout if timeout else self._request_timeout

        now = time.time()

        href = self.pick_http_href(td, td.get_action_forms(name))

        if href is None:
            raise FormNotFoundException()

        body = json.dumps({"input": input_value})
        http_client = tornado.httpclient.AsyncHTTPClient()

        try:
            http_request = tornado.httpclient.HTTPRequest(
                href, method="POST",
                body=body,
                headers=self.JSON_HEADERS,
                connect_timeout=con_timeout,
                request_timeout=req_timeout)
        except HTTPTimeoutError:
            raise ClientRequestTimeout

        response = yield http_client.fetch(http_request)
        invocation_url = json.loads(response.body).get("invocation")

        @tornado.gen.coroutine
        def check_invocation():
            parsed = parse.urlparse(href)

            invoc_href = "{}://{}/{}".format(
                parsed.scheme,
                parsed.netloc,
                invocation_url.lstrip("/"))

            invoc_http_req = tornado.httpclient.HTTPRequest(
                invoc_href, method="GET",
                connect_timeout=con_timeout,
                request_timeout=req_timeout)

            self._logr.debug("Checking invocation: {}".format(invocation_url))

            try:
                invoc_res = yield http_client.fetch(invoc_http_req)
            except HTTPTimeoutError:
                self._logr.debug("Timeout checking invocation: {}".format(invocation_url))
                raise tornado.gen.Return((False, None))

            status = json.loads(invoc_res.body)

            if status.get("done") is False:
                raise tornado.gen.Return((False, None))

            if status.get("error") is not None:
                raise tornado.gen.Return((True, Exception(status.get("error"))))
            else:
                raise tornado.gen.Return((True, tornado.gen.Return(status.get("result"))))

        while True:
            done, result = yield check_invocation()

            if done:
                raise result
            elif timeout and (time.time() - now) > timeout:
                raise ClientRequestTimeout

    @tornado.gen.coroutine
    def write_property(self, td, name, value, timeout=None):
        """Updates the value of a Property on a remote Thing.
        Returns a Future."""

        con_timeout = timeout if timeout else self._connect_timeout
        req_timeout = timeout if timeout else self._request_timeout

        href = self.pick_http_href(td, td.get_property_forms(name))

        if href is None:
            raise FormNotFoundException()

        http_client = tornado.httpclient.AsyncHTTPClient()
        body = json.dumps({"value": value})

        try:
            http_request = tornado.httpclient.HTTPRequest(
                href, method="PUT", body=body,
                headers=self.JSON_HEADERS,
                connect_timeout=con_timeout,
                request_timeout=req_timeout)
        except HTTPTimeoutError:
            raise ClientRequestTimeout

        yield http_client.fetch(http_request)

    @tornado.gen.coroutine
    def read_property(self, td, name, timeout=None):
        """Reads the value of a Property on a remote Thing.
        Returns a Future."""

        con_timeout = timeout if timeout else self._connect_timeout
        req_timeout = timeout if timeout else self._request_timeout

        href = self.pick_http_href(td, td.get_property_forms(name))

        if href is None:
            raise FormNotFoundException()

        http_client = tornado.httpclient.AsyncHTTPClient()

        try:
            http_request = tornado.httpclient.HTTPRequest(
                href, method="GET",
                connect_timeout=con_timeout,
                request_timeout=req_timeout)
        except HTTPTimeoutError:
            raise ClientRequestTimeout

        response = yield http_client.fetch(http_request)

        raise tornado.gen.Return(json.loads(response.body).get("value"))

    def on_event(self, td, name):
        """Subscribes to an event on a remote Thing.
        Returns an Observable."""

        href = self.pick_http_href(td, td.get_event_forms(name))

        if href is None:
            raise FormNotFoundException()

        def subscribe(observer):
            """Subscription function to observe events using the HTTP protocol."""

            state = {"active": True}

            @handle_observer_finalization(observer)
            @tornado.gen.coroutine
            def callback():
                http_client = tornado.httpclient.AsyncHTTPClient()
                http_request = tornado.httpclient.HTTPRequest(href, method="GET")

                while state["active"]:
                    try:
                        response = yield http_client.fetch(http_request)
                        payload = json.loads(response.body).get("payload")
                        observer.on_next(EmittedEvent(init=payload, name=name))
                    except HTTPTimeoutError:
                        pass

            def unsubscribe():
                state["active"] = False

            tornado.ioloop.IOLoop.current().add_callback(callback)

            return unsubscribe

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_property_change(self, td, name):
        """Subscribes to property changes on a remote Thing.
        Returns an Observable"""

        href = self.pick_http_href(td, td.get_property_forms(name), op=InteractionVerbs.OBSERVE_PROPERTY)

        if href is None:
            raise FormNotFoundException()

        def subscribe(observer):
            """Subscription function to observe property updates using the HTTP protocol."""

            state = {"active": True}

            @handle_observer_finalization(observer)
            @tornado.gen.coroutine
            def callback():
                http_client = tornado.httpclient.AsyncHTTPClient()
                http_request = tornado.httpclient.HTTPRequest(href, method="GET")

                while state["active"]:
                    try:
                        response = yield http_client.fetch(http_request)
                        value = json.loads(response.body).get("value")
                        init = PropertyChangeEventInit(name=name, value=value)
                        observer.on_next(PropertyChangeEmittedEvent(init=init))
                    except HTTPTimeoutError:
                        pass

            def unsubscribe():
                state["active"] = False

            tornado.ioloop.IOLoop.current().add_callback(callback)

            return unsubscribe

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_td_change(self, url):
        """Subscribes to Thing Description changes on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents a Thing consumed by a servient.
"""

import tornado.gen
from rx.concurrency import IOLoopScheduler

from wotpy.wot.consumed.interaction_map import \
    ConsumedThingPropertyDict, \
    ConsumedThingActionDict, \
    ConsumedThingEventDict


class ConsumedThing(object):
    """An entity that serves to interact with a Thing.
    An application uses this class when it acts as a *client* of the Thing."""

    def __init__(self, servient, td):
        self._servient = servient
        self._td = td

    def __str__(self):
        return "<{}> {}".format(self.__class__.__name__, self.td.id)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private ThingFragment instance before propagating the exception."""

        return getattr(self.td.to_thing_fragment(), name)

    @property
    def servient(self):
        """Returns the Servient that contains this Consumed Thing."""

        return self._servient

    @property
    def td(self):
        """Returns the ThingDescription instance that represents
        the TD that this Consumed Thing is based on."""

        return self._td

    @tornado.gen.coroutine
    def invoke_action(self, name, input_value=None, timeout=None, client_kwargs=None):
        """Takes the Action name from the name argument and the list of parameters,
        then requests from the underlying platform and the Protocol Bindings to invoke
        the Action on the remote Thing and return the result.
        Returns a Future that resolves with the return value or rejects with an Error."""

        client = self.servient.select_client(self.td, name)
        client_kwargs = client_kwargs if client_kwargs else {}

        result = yield client.invoke_action(
            self.td, name, input_value,
            timeout=timeout,
            **client_kwargs.get(client.protocol, {}))

        raise tornado.gen.Return(result)

    @tornado.gen.coroutine
    def write_property(self, name, value, timeout=None, client_kwargs=None):
        """Takes the Property name as the name argument and the new value as the value
        argument, then requests from the underlying platform and the Protocol Bindings
        to update the Property on the remote Thing and return the result.
        Returns a Future that resolves on success or rejects with an Error."""

        client = self.servient.select_client(self.td, name)
        client_kwargs = client_kwargs if client_kwargs else {}

        yield client.write_property(
            self.td, name, value,
            timeout=timeout,
            **client_kwargs.get(client.protocol, {}))

    @tornado.gen.coroutine
    def read_property(self, name, timeout=None, client_kwargs=None):
        """Takes the Property name as the name argument, then requests from the
        underlying platform and the Protocol Bindings to retrieve the Property
        on the remote Thing and return the result.
        Returns a Future that resolves with the Property value or rejects with an Error."""

        client = self.servient.select_client(self.td, name)
        client_kwargs = client_kwargs if client_kwargs else {}

        value = yield client.read_property(
            self.td, name,
            timeout=timeout,
            **client_kwargs.get(client.protocol, {}))

        raise tornado.gen.Return(value)

    def on_event(self, name, client_kwargs=None):
        """Returns an Observable for the Event specified in the name argument,
        allowing subscribing to and unsubscribing from notifications."""

        client = self.servient.select_client(self.td, name)
        client_kwargs = client_kwargs if client_kwargs else {}

        return client.on_event(
            self.td, name,
            **client_kwargs.get(client.protocol, {}))

    def on_property_change(self, name, client_kwargs=None):
        """Returns an Observable for the Property specified in the name argument,
        allowing subscribing to and unsubscribing from notifications."""

        client = self.servient.select_client(self.td, name)
        client_kwargs = client_kwargs if client_kwargs else {}

        return client.on_property_change(
            self.td, name,
            **client_kwargs.get(client.protocol, {}))

    def on_td_change(self):
        """Returns an Observable, allowing subscribing to and unsubscribing
        from notifications to the Thing Description."""

        raise NotImplementedError()

    @property
    def properties(self):
        """Returns a dictionary of ThingProperty items."""

        return ConsumedThingPropertyDict(consumed_thing=self)

    @property
    def actions(self):
        """Returns a dictionary of ThingAction items."""

        return ConsumedThingActionDict(consumed_thing=self)

    @property
    def events(self):
        """Returns a dictionary of ThingEvent items."""

        return ConsumedThingEventDict(consumed_thing=self)

    @property
    def links(self):
        """Represents a dictionary of WebLink items."""

        raise NotImplementedError()

    def subscribe(self, *args, **kwargs):
        """Subscribes to changes on the TD of this thing."""

        observable = self.on_td_change()
        return observable.subscribe_on(IOLoopScheduler()).subscribe(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper class for dictionaries to represent Things.
"""

import six

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict, EventFragmentDict
from wotpy.wot.dictionaries.link import LinkDict
from wotpy.wot.dictionaries.security import SecuritySchemeDict
from wotpy.utils.utils import to_camel
from wotpy.wot.dictionaries.version import VersioningDict
from wotpy.wot.enums import SecuritySchemeType


class ThingFragment(WotBaseDict):
    """ThingFragment is a wrapper around a dictionary that contains properties
    representing semantic metadata and interactions (Properties, Actions and Events).
    It is used for initializing an internal representation of a Thing Description,
    and it is also used in ThingFilter."""

    class Meta:
        fields = {
            "id",
            "version",
            "name",
            "description",
            "support",
            "created",
            "lastModified",
            "base",
            "properties",
            "actions",
            "events",
            "links",
            "security"
        }

        required = {
            "id"
        }

        fields_readonly = [
            "id"
        ]

        fields_str = [
            "name",
            "description",
            "support",
            "created",
            "lastModified",
            "base"
        ]

        fields_dict = [
            "properties",
            "actions",
            "events"
        ]

        fields_list = [
            "links",
            "security"
        ]

        fields_instance = [
            "version"
        ]

        assert set(fields_readonly + fields_str + fields_dict + fields_list + fields_instance) == fields

    def __setattr__(self, name, value):
        """Checks to see if the attribute that is being set is a
        Thing fragment property and updates the internal dict."""

        name_camel = to_camel(name)

        if name_camel not in self.Meta.fields:
            return super(ThingFragment, self).__setattr__(name, value)

        if name_camel in self.Meta.fields_readonly:
            raise AttributeError("Can't set attribute {}".format(name))

        if name_camel in self.Meta.fields_str:
            self._init[name_camel] = value
            return

        if name_camel in self.Meta.fields_dict:
            self._init[name_camel] = {key: val.to_dict() for key, val in six.iteritems(value)}
            return

        if name_camel in self.Meta.fields_list:
            self._init[name_camel] = [item.to_dict() for item in value]
            return

        if name_camel in self.Meta.fields_instance:
            self._init[name_camel] = value.to_dict()
            return

    @property
    def name(self):
        """The name of the Thing.
        This property returns the ID if the name is undefined."""

        return self._init.get("name", self.id)

    @property
    def security(self):
        """Set of security configurations, provided as an array,
        that must all be satisfied for access to resources at or
        below the current level, if not overridden at a lower level.
        A default nosec security scheme will be provided if none are defined."""

        if "security" not in self._init:
            return [SecuritySchemeDict.build({"scheme": SecuritySchemeType.NOSEC})]

        return [SecuritySchemeDict.build(item) for item in self._init.get("security")]

    @property
    def properties(self):
        """The properties optional attribute represents a dict with keys
        that correspond to Property names and values of type PropertyFragment."""

        return {
            key: PropertyFragmentDict(val)
            for key, val in six.iteritems(self._init.get("properties", {}))
        }

    @property
    def actions(self):
        """The actions optional attribute represents a dict with keys
        that correspond to Action names and values of type ActionFragment."""

        return {
            key: ActionFragmentDict(val)
            for key, val in six.iteritems(self._init.get("actions", {}))
        }

    @property
    def events(self):
        """The events optional attribute represents a dictionary with keys
        that correspond to Event names and values of type EventFragment."""

        return {
            key: EventFragmentDict(val)
            for key, val in six.iteritems(self._init.get("events", {}))
        }

    @property
    def links(self):
        """The links optional attribute represents an array of Link objects."""

        return [LinkDict(item) for item in self._init.get("links", [])]

    @property
    def version(self):
        """Provides version information."""

        return VersioningDict(self._init.get("version")) if self._init.get("version") else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base class for WoT dictionaries.
"""

import six

from wotpy.utils.utils import merge_args_kwargs_dict, to_camel, to_snake


class WotBaseDict(object):
    """Base class for all WoT data types represented
    as dictionaries in the Scripting API specification."""

    class Meta:
        fields = set()
        required = set()
        defaults = dict()

    def __init__(self, *args, **kwargs):
        """Constructor.
        Will raise ValueError if there is some required field missing."""

        init_dict = merge_args_kwargs_dict(args, kwargs)

        self._init = {}

        for key, val in six.iteritems(init_dict):
            self._init.update({to_camel(key): val})

        try:
            required = self.Meta.required
        except AttributeError:
            required = []

        for field in required:
            if field not in self._init:
                raise ValueError("Missing required field: {}".format(field))

    def __getattr__(self, name):
        """Transforms the field name to camelCase and
        attemps to retrieve it from the internal dict."""

        name_camel = to_camel(name)

        if name_camel not in self.Meta.fields:
            raise AttributeError(name)

        if name_camel in self._init:
            return self._init[name_camel]

        try:
            return self.Meta.defaults.get(name_camel, None)
        except AttributeError:
            return None

    def to_dict(self):
        """Returns the pure dict (JSON-serializable) representation of this WoT dictionary."""

        ret = {}

        def is_list_wot_dicts(x):
            return isinstance(x, list) and len(x) and hasattr(x[0], "to_dict")

        def is_dict_wot_dicts(x):
            return isinstance(x, dict) and len(x) and hasattr(next(six.itervalues(x)), "to_dict")

        def is_wot_dict(x):
            return hasattr(x, "to_dict")

        existing_fields = [
            f for f in self.Meta.fields
            if f in self._init or (to_snake(f) in dir(self) and getattr(self, to_snake(f)) is not None)
        ]

        for name_camel in existing_fields:
            field_val = getattr(self, to_snake(name_camel))

            if is_list_wot_dicts(field_val):
                field_val = [item.to_dict() for item in field_val]
            elif is_dict_wot_dicts(field_val):
                field_val = {key: val.to_dict() for key, val in six.iteritems(field_val)}
            elif is_wot_dict(field_val):
                field_val = field_val.to_dict()

            ret.update({name_camel: field_val})

        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os

import tornado.gen
import tornado.ioloop
from hbmqtt.client import MQTTClient, ConnectException

from wotpy.protocols.mqtt.enums import MQTTCodesACK

ENV_BROKER_URL = "WOTPY_TESTS_MQTT_BROKER_URL"
BROKER_SKIP_REASON = "The test MQTT broker is offline"


def get_test_broker_url():
    """Returns the MQTT broker URL defined in the environment."""

    return os.environ.get(ENV_BROKER_URL, None)


def is_test_broker_online():
    """Returns True if the MQTT broker defined in the environment is online."""

    @tornado.gen.coroutine
    def check_conn():
        broker_url = get_test_broker_url()

        if not broker_url:
            logging.warning("Undefined MQTT broker URL")
            raise tornado.gen.Return(False)

        try:
            hbmqtt_client = MQTTClient()
            ack_con = yield hbmqtt_client.connect(broker_url)
            if ack_con != MQTTCodesACK.CON_OK:
                logging.warning("Error ACK on MQTT broker connection: {}".format(ack_con))
                raise tornado.gen.Return(False)
        except ConnectException as ex:
            logging.warning("MQTT broker connection error: {}".format(ex))
            raise tornado.gen.Return(False)

        raise tornado.gen.Return(True)

    conn_ok = tornado.ioloop.IOLoop.current().run_sync(check_conn)

    if conn_ok is False:
        logging.warning(
            "Couldn't connect to the test MQTT broker. "
            "Please check the {} variable".format(ENV_BROKER_URL))

    return conn_ok



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from os import path

from setuptools import setup, find_packages

from wotpy.__version__ import __version__
from wotpy.support import is_dnssd_supported, is_coap_supported, is_mqtt_supported

install_requires = [
    'tornado>=5.1,<6.0',
    'jsonschema>=2.0,<3.0',
    'six>=1.10.0,<2.0',
    'rx>=1.6.0,<2.0',
    'python-slugify>=1.2.4,<2.0'
]

test_requires = [
    'pytest>=3.6.1,<4.0.0',
    'pytest-cov>=2.5.1,<2.6.0',
    'pytest-rerunfailures>=4.1,<5.0',
    'mock>=2.0,<3.0',
    'tox>=2.0,<3.0',
    'faker>=0.8.15,<0.9',
    'Sphinx>=1.7.5,<2.0.0',
    'sphinx-rtd-theme>=0.4.0,<0.5.0',
    'futures>=3.1.1,<4.0.0',
    'pyOpenSSL>=18.0.0,<19.0.0',
    'python-coveralls==2.9.1',
    'coverage==4.0.3'
]

if is_coap_supported():
    install_requires.append('aiocoap[linkheader]==0.4a1')

if is_mqtt_supported():
    install_requires.append('hbmqtt>=0.9.4,<1.0')

if is_dnssd_supported():
    install_requires.append('zeroconf>=0.21.3,<0.22.0')
    test_requires.append('aiozeroconf==0.1.8')

this_dir = path.abspath(path.dirname(__file__))

with open(path.join(this_dir, 'README.md')) as fh:
    long_description = fh.read()

setup(
    name='wotpy',
    version=__version__,
    description='Python implementation of a W3C WoT Runtime and the WoT Scripting API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='wot iot gateway fog w3c',
    author='Andres Garcia Mangas',
    author_email='andres.garcia@fundacionctic.org',
    url='https://github.com/agmangas/wot-py',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'tests': test_requires,
        'uvloop': ['uvloop>=0.12.2,<0.13.0']
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that subscribes to all the Interactions of one kind for
all the ExposedThings contained by a Protocol Binding server.
"""

import logging

import six

from wotpy.wot.enums import InteractionTypes


class InteractionsSubscriber(object):
    """Class that subscribes to all the Interactions of one kind for
    all the ExposedThings contained by a Protocol Binding server."""

    def __init__(self, interaction_type, server, on_next_builder):
        assert interaction_type in [InteractionTypes.PROPERTY, InteractionTypes.EVENT]
        self._interaction_type = interaction_type
        self._server = server
        self._on_next_builder = on_next_builder
        self._subs = {}
        self._logr = logging.getLogger(__name__)

    def _dispose_exposed_thing_subs(self, exp_thing):
        """Disposes of all currently active subscriptions for the given ExposedThing."""

        if exp_thing not in self._subs:
            return

        for key in self._subs[exp_thing]:
            self._subs[exp_thing][key].dispose()

        self._subs.pop(exp_thing)

    def _interaction_attr_name(self):
        """Returns the attribute name of the Thing and ExposedThing
        iterator for the current type of interactions."""

        return {
            InteractionTypes.PROPERTY: "properties",
            InteractionTypes.EVENT: "events"
        }.get(self._interaction_type)

    def _get_exposed_thing_interaction_set(self, exp_thing):
        """Returns the set of interactions that should be observed."""

        attr = self._interaction_attr_name()

        intrc_expected = set(six.itervalues(exp_thing.thing.__getattribute__(attr)))

        if self._interaction_type == InteractionTypes.PROPERTY:
            intrc_expected = set(item for item in intrc_expected if item.observable)

        return intrc_expected

    def _refresh_exposed_thing_subs(self, exp_thing):
        """Refresh the subscriptions for the given ExposedThing."""

        if exp_thing not in self._subs:
            self._subs[exp_thing] = {}

        thing_subs = self._subs[exp_thing]

        intrc_expected = self._get_exposed_thing_interaction_set(exp_thing)
        intrc_current = set(thing_subs.keys())
        intrc_remove = intrc_current.difference(intrc_expected)

        for intrc in intrc_remove:
            thing_subs[intrc].dispose()
            thing_subs.pop(intrc)

        intrc_new = [item for item in intrc_expected if item not in thing_subs]

        attr = self._interaction_attr_name()

        for intrc in intrc_new:
            on_next = self._on_next_builder(exp_thing, intrc)
            exp_thing_intrc = exp_thing.__getattribute__(attr)[intrc.name]

            def on_error(err):
                self._logr.warning("Error on subscription to {}: {}".format(exp_thing_intrc, err))
                thing_subs[intrc].dispose()
                thing_subs.pop(intrc)

            thing_subs[intrc] = exp_thing_intrc.subscribe(on_next=on_next, on_error=on_error)

    def dispose(self):
        """Disposes of all the currently active subscriptions."""

        for exp_thing in list(six.iterkeys(self._subs)):
            self._dispose_exposed_thing_subs(exp_thing)

    def refresh(self):
        """Refresh all subscriptions for the entire set of ExposedThings."""

        things_expected = set(self._server.exposed_things)
        things_current = set(self._subs.keys())
        things_remove = things_current.difference(things_expected)

        for exp_thing in things_remove:
            self._dispose_exposed_thing_subs(exp_thing)

        for exp_thing in things_expected:
            self._refresh_exposed_thing_subs(exp_thing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility functions used by client and server implementations.
"""

from six.moves import urllib


def is_scheme_form(form, base, scheme):
    """Returns True if the scheme of the URI for
    the given Form matches the scheme argument."""

    resolved_url = form.resolve_uri(base=base)

    if not resolved_url:
        return False

    parsed_scheme = urllib.parse.urlparse(resolved_url).scheme

    return parsed_scheme in scheme if isinstance(scheme, list) else parsed_scheme == scheme


def pick_form(td, forms, schemes, op=None):
    """Picks the Form that will be used to connect to the remote Thing."""

    for scheme in schemes:
        scheme_forms = [
            form for form in forms
            if is_scheme_form(form, td.base, scheme)
        ]

        if op is not None:
            scheme_forms = [form for form in scheme_forms if form.op == op]

        if len(scheme_forms):
            return scheme_forms[0]

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
ExposedThing and related entities.

.. autosummary::
    :toctree: _exposed

    wotpy.wot.exposed.interaction_map
    wotpy.wot.exposed.thing
    wotpy.wot.exposed.thing_set
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents the abstract client interface.
"""

from abc import ABCMeta, abstractmethod


class BaseProtocolClient(object):
    """Base protocol client class.
    This is the interface that must be implemented by all client classes."""

    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def protocol(self):
        """Protocol of this client instance.
        A member of the Protocols enum."""

        raise NotImplementedError()

    @abstractmethod
    def is_supported_interaction(self, td, name):
        """Returns True if the any of the Forms for the Interaction
        with the given name is supported in this Protocol Binding client."""

        raise NotImplementedError()

    @abstractmethod
    def invoke_action(self, td, name, input_value, timeout=None):
        """Invokes an Action on a remote Thing.
        Returns a Future."""

        raise NotImplementedError()

    @abstractmethod
    def write_property(self, td, name, value, timeout=None):
        """Updates the value of a Property on a remote Thing.
        Returns a Future."""

        raise NotImplementedError()

    @abstractmethod
    def read_property(self, td, name, timeout=None):
        """Reads the value of a Property on a remote Thing.
        Returns a Future."""

        raise NotImplementedError()

    @abstractmethod
    def on_event(self, td, name):
        """Subscribes to an event on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError()

    @abstractmethod
    def on_property_change(self, td, name):
        """Subscribes to property changes on a remote Thing.
        Returns an Observable"""

        raise NotImplementedError()

    @abstractmethod
    def on_td_change(self, url):
        """Subscribes to Thing Description changes on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents a WoT servient.
"""

import functools
import socket

import six
import tornado.concurrent
import tornado.gen
import tornado.ioloop
import tornado.locks
import tornado.web

from wotpy.protocols.enums import Protocols
from wotpy.protocols.http.client import HTTPClient
from wotpy.protocols.ws.client import WebsocketClient
from wotpy.support import is_coap_supported, is_mqtt_supported, is_dnssd_supported
from wotpy.utils.utils import get_main_ipv4_address
from wotpy.wot.enums import InteractionTypes
from wotpy.wot.exposed.thing_set import ExposedThingSet
from wotpy.wot.td import ThingDescription
from wotpy.wot.wot import WoT


# noinspection PyAbstractClass,PyMethodOverriding,PyAttributeOutsideInit
class TDHandler(tornado.web.RequestHandler):
    """Handler that returns the TD document of a given Thing."""

    def initialize(self, servient):
        self.servient = servient

    def get(self, thing_url_name):
        exp_thing = self.servient.exposed_thing_set.find_by_thing_id(thing_url_name)

        td_doc = ThingDescription.from_thing(exp_thing.thing).to_dict()
        base_url = self.servient.get_thing_base_url(exp_thing)

        if base_url:
            td_doc.update({"base": base_url})

        self.write(td_doc)


# noinspection PyAbstractClass,PyMethodOverriding,PyAttributeOutsideInit
class TDCatalogueHandler(tornado.web.RequestHandler):
    """Handler that returns the entire catalogue of Things contained in this servient.
    May return TDs in expanded format or URL pointers to the individual TDs."""

    def initialize(self, servient):
        self.servient = servient

    def get(self):
        response = {}

        for exp_thing in self.servient.enabled_exposed_things:
            thing_id = exp_thing.thing.id

            if self.get_argument("expanded", False):
                val = ThingDescription.from_thing(exp_thing.thing).to_dict()
                val.update({"base": self.servient.get_thing_base_url(exp_thing)})
            else:
                val = "/{}".format(exp_thing.thing.url_name)

            response[thing_id] = val

        self.write(response)


class ServientStateException(Exception):
    """Exception raised when the user modifies the Servient while
    the Servient is in an inappropriate state."""

    pass


def _stopped_servient_only(func):
    """Decorator that raises an Exception when attempting
    to call the function while the Servient is running."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        servient = args[0]

        if servient.is_running:
            raise ServientStateException("Attempted to modify the Servient while it was running")

        return func(*args, **kwargs)

    return wrapper


def _get_hostname_fallback():
    """Tries to guess the hostname of the current host that should be used on TD Forms.
    Two strategies are used for this: First, the socket.getfqdn() method. If the returned
    value is not a FQDN then we try to get the IPv4 address of the main network interface."""

    fqdn = socket.getfqdn()

    return fqdn if "." in fqdn else get_main_ipv4_address()


class Servient(object):
    """An entity that is both a WoT client and server at the same time.
    WoT servers are Web servers that possess capabilities to access underlying
    IoT devices and expose a public interface named the WoT Interface that may
    be used by other clients.
    WoT clients are entities that are able to understand the WoT Interface to
    send requests and interact with IoT devices exposed by other WoT servients
    or servers using the capabilities of a Web client such as Web browser."""

    def __init__(self, hostname=None, catalogue_port=9090,
                 clients=None, clients_config=None,
                 dnssd_enabled=False, dnssd_instance_name=None):
        self._hostname = hostname if hostname is not None else _get_hostname_fallback()

        if not isinstance(self._hostname, six.string_types):
            raise ValueError("Invalid hostname: {}".format(self._hostname))

        if isinstance(clients, list):
            clients = {item.protocol: item for item in clients}

        self._servers = {}
        self._clients = clients if clients else {}
        self._clients_config = clients_config
        self._catalogue_port = catalogue_port
        self._catalogue_server = None
        self._exposed_thing_set = ExposedThingSet()
        self._servient_lock = tornado.locks.Lock()
        self._is_running = False
        self._dnssd_enabled = dnssd_enabled if dnssd_enabled and is_dnssd_supported() else False
        self._dnssd_instance_name = dnssd_instance_name
        self._dnssd = None
        self._enabled_exposed_thing_ids = set()

        if not len(self._clients):
            self._build_default_clients()

    @staticmethod
    def _default_select_client(clients, td, name):
        """Default implementation of the function to select
        a Protocol Binding client for an Interaction."""

        protocol_preference_map = {
            InteractionTypes.PROPERTY: [
                Protocols.HTTP,
                Protocols.COAP,
                Protocols.WEBSOCKETS,
                Protocols.MQTT
            ],
            InteractionTypes.ACTION: [
                Protocols.WEBSOCKETS,
                Protocols.MQTT,
                Protocols.COAP,
                Protocols.HTTP
            ],
            InteractionTypes.EVENT: [
                Protocols.WEBSOCKETS,
                Protocols.MQTT,
                Protocols.COAP,
                Protocols.HTTP
            ]
        }

        supported_protocols = [
            client.protocol for client in clients
            if client.is_supported_interaction(td, name)
        ]

        intrct_names = {
            InteractionTypes.PROPERTY: six.iterkeys(td.properties),
            InteractionTypes.ACTION: six.iterkeys(td.actions),
            InteractionTypes.EVENT: six.iterkeys(td.events)
        }

        try:
            intrct_type = next(key for key, names in six.iteritems(intrct_names) if name in names)
        except StopIteration:
            raise ValueError("Unknown interaction: {}".format(name))

        protocol_prefs = protocol_preference_map[intrct_type]
        protocol_choices = set(protocol_prefs).intersection(set(supported_protocols))

        if not len(protocol_choices):
            return list(clients)[0]

        protocol = next(proto for proto in protocol_prefs if proto in protocol_choices)

        return next(client for client in clients if client.protocol == protocol)

    @property
    def is_running(self):
        """Returns True if the Servient is currently running
        (i.e. the attached servers have been started)."""

        return self._is_running

    @property
    def hostname(self):
        """Hostname attached to this servient."""

        return self._hostname

    @property
    def exposed_thing_set(self):
        """Returns the ExposedThingSet instance that
        contains the ExposedThings of this servient."""

        return self._exposed_thing_set

    @property
    def exposed_things(self):
        """Returns an iterator for the ExposedThings contained in this Servient."""

        return self.exposed_thing_set.exposed_things

    @property
    def enabled_exposed_things(self):
        """Returns an iterator for the enabled ExposedThings contained in this Servient."""

        for exposed_thing in self.exposed_things:
            if exposed_thing.id in self._enabled_exposed_thing_ids:
                yield exposed_thing

    @property
    def servers(self):
        """Returns the dict of Protocol Binding servers attached to this servient."""

        return self._servers

    @property
    def clients(self):
        """Returns the dict of Protocol Binding clients attached to this servient."""

        return self._clients

    @property
    def catalogue_port(self):
        """Returns the current port of the HTTP Thing Description catalogue service."""

        return self._catalogue_port

    @catalogue_port.setter
    @_stopped_servient_only
    def catalogue_port(self, port):
        """Enables the servient TD catalogue in the given port."""

        self._catalogue_port = port

    @property
    def dnssd(self):
        """Returns the DNS-SD instance linked to this Servient (if enabled and started)."""

        return self._dnssd

    @property
    def dnssd_instance_name(self):
        """Returns the user-given DNS-SD service instance name."""

        return self._dnssd_instance_name

    @tornado.gen.coroutine
    def _start_dnssd(self):
        """Starts the DNS-SD service and registers the servient."""

        if self._dnssd or not self._dnssd_enabled:
            return

        from wotpy.wot.discovery.dnssd.service import DNSSDDiscoveryService

        self._dnssd = DNSSDDiscoveryService()

        yield self._dnssd.start()
        yield self._dnssd.register(self, instance_name=self._dnssd_instance_name)

    @tornado.gen.coroutine
    def _stop_dnssd(self):
        """Unregisters the servient and stops the DNS-SD service."""

        if not self._dnssd:
            return

        yield self._dnssd.stop()

        self._dnssd = None

    # noinspection PyArgumentList
    def _build_default_clients(self):
        """Builds the default Protocol Binding clients."""

        self._clients = self._clients if self._clients else {}

        conf = self._clients_config if self._clients_config else {}

        self._clients.update({
            Protocols.WEBSOCKETS: WebsocketClient(**conf.get(Protocols.WEBSOCKETS, {})),
            Protocols.HTTP: HTTPClient(**conf.get(Protocols.HTTP, {}))
        })

        if is_coap_supported():
            from wotpy.protocols.coap.client import CoAPClient
            self._clients.update({Protocols.COAP: CoAPClient(**conf.get(Protocols.COAP, {}))})

        if is_mqtt_supported():
            from wotpy.protocols.mqtt.client import MQTTClient
            self._clients.update({Protocols.MQTT: MQTTClient(**conf.get(Protocols.MQTT, {}))})

    def _build_td_catalogue_app(self):
        """Returns a Tornado app that provides one endpoint to retrieve the
        entire catalogue of thing descriptions contained in this servient."""

        return tornado.web.Application([
            (r"/", TDCatalogueHandler, dict(servient=self)),
            (r"/(?P<thing_url_name>[^\/]+)", TDHandler, dict(servient=self))
        ])

    def _start_catalogue(self):
        """Starts the TD catalogue server if enabled."""

        if self._catalogue_server or not self._catalogue_port:
            return

        catalogue_app = self._build_td_catalogue_app()
        self._catalogue_server = catalogue_app.listen(self._catalogue_port)

    def _stop_catalogue(self):
        """Stops the TD catalogue server if running."""

        if not self._catalogue_server:
            return

        self._catalogue_server.stop()
        self._catalogue_server = None

    def _clean_forms(self):
        """Cleans all the Forms from all the ExposedThings contained in this Servient."""

        for exposed_thing in self._exposed_thing_set.exposed_things:
            for interaction in exposed_thing.thing.interactions:
                interaction.clean_forms()

    def _clean_protocol_forms(self, exposed_thing, protocol):
        """Removes all interaction forms linked to this
        server protocol for the given ExposedThing."""

        assert self._exposed_thing_set.contains(exposed_thing)
        assert protocol in self._servers

        for interaction in exposed_thing.thing.interactions:
            forms_to_remove = [
                form for form in interaction.forms
                if form.protocol == protocol
            ]

            for form in forms_to_remove:
                interaction.remove_form(form)

    def _server_has_exposed_thing(self, server, exposed_thing):
        """Returns True if the given server contains the ExposedThing."""

        assert server in self._servers.values()
        assert self._exposed_thing_set.contains(exposed_thing)

        return server.exposed_thing_set.contains(exposed_thing)

    def _add_interaction_forms(self, server, exposed_thing):
        """Builds and adds to the ExposedThing the Links related to the given server."""

        assert server in self._servers.values()
        assert self._exposed_thing_set.contains(exposed_thing)

        for interaction in exposed_thing.thing.interactions:
            forms = server.build_forms(hostname=self._hostname, interaction=interaction)

            for form in forms:
                interaction.add_form(form)

    def _regenerate_server_forms(self, server):
        """Cleans and regenerates Forms for the given server in all ExposedThings."""

        assert server in self._servers.values()

        for exp_thing in self._exposed_thing_set.exposed_things:
            self._clean_protocol_forms(exp_thing, server.protocol)
            if self._server_has_exposed_thing(server, exp_thing):
                self._add_interaction_forms(server, exp_thing)

    def get_thing_base_url(self, exposed_thing):
        """Return the base URL for the given ExposedThing
        for one of the currently active servers."""

        if exposed_thing.thing.base:
            return exposed_thing.base

        if not self.exposed_thing_set.contains(exposed_thing):
            raise ValueError("Unknown ExposedThing")

        if not len(self.servers):
            return None

        protocol_default = sorted(six.iterkeys(self.servers))[0]
        protocol = Protocols.HTTP if Protocols.HTTP in self.servers else protocol_default
        server = self.servers[protocol]

        return server.build_base_url(hostname=self.hostname, thing=exposed_thing.thing)

    def select_client(self, td, name):
        """Returns the Protocol Binding client instance to
        communicate with the given Interaction."""

        return Servient._default_select_client(self.clients.values(), td, name)

    @_stopped_servient_only
    def add_client(self, client):
        """Adds a new Protocol Binding client to this servient."""

        self._clients[client.protocol] = client

    @_stopped_servient_only
    def remove_client(self, protocol):
        """Removes the Protocol Binding client with the given protocol from this servient."""

        self._clients.pop(protocol, None)

    @_stopped_servient_only
    def add_server(self, server):
        """Adds a new Protocol Binding server to this servient."""

        self._servers[server.protocol] = server

    @_stopped_servient_only
    def remove_server(self, protocol):
        """Removes the Protocol Binding server with the given protocol from this servient."""

        self._servers.pop(protocol, None)

    def refresh_forms(self):
        """Cleans and regenerates Forms for all the
        ExposedThings and servers contained in this servient."""

        self._clean_forms()

        for server in self._servers.values():
            self._regenerate_server_forms(server)

    def enable_exposed_thing(self, thing_id):
        """Enables the ExposedThing with the given ID.
        This is, the servers will listen for requests for this thing."""

        exposed_thing = self.get_exposed_thing(thing_id)

        for server in self._servers.values():
            server.add_exposed_thing(exposed_thing)
            self._regenerate_server_forms(server)

        self._enabled_exposed_thing_ids.add(exposed_thing.id)

    def disable_exposed_thing(self, thing_id):
        """Disables the ExposedThing with the given ID.
        This is, the servers will not listen for requests for this thing."""

        exposed_thing = self.get_exposed_thing(thing_id)

        if exposed_thing.id not in self._enabled_exposed_thing_ids:
            raise ValueError("ExposedThing {} is already disabled".format(thing_id))

        for server in self._servers.values():
            server.remove_exposed_thing(exposed_thing.id)
            self._regenerate_server_forms(server)

        self._enabled_exposed_thing_ids.remove(exposed_thing.id)

    def add_exposed_thing(self, exposed_thing):
        """Adds an ExposedThing to this Servient.
        ExposedThings are disabled by default."""

        self._exposed_thing_set.add(exposed_thing)

    def remove_exposed_thing(self, thing_id):
        """Disables and removes an ExposedThing from this Servient."""

        if thing_id in self._enabled_exposed_thing_ids:
            self.disable_exposed_thing(thing_id)

        self._exposed_thing_set.remove(thing_id)

    def get_exposed_thing(self, thing_id):
        """Finds and returns an ExposedThing contained in this servient by Thing ID.
        Raises ValueError if the ExposedThing is not present."""

        exp_thing = self._exposed_thing_set.find_by_thing_id(thing_id)

        if exp_thing is None:
            raise ValueError("Unknown ExposedThing: {}".format(thing_id))

        return exp_thing

    @_stopped_servient_only
    def disable_td_catalogue(self):
        """Disables the servient TD catalogue."""

        self._catalogue_port = None

    @tornado.gen.coroutine
    def start(self):
        """Starts the servers and returns an instance of the WoT object."""

        with (yield self._servient_lock.acquire()):
            self.refresh_forms()
            yield [server.start() for server in six.itervalues(self._servers)]
            self._start_catalogue()
            yield self._start_dnssd()
            self._is_running = True

            raise tornado.gen.Return(WoT(servient=self))

    @tornado.gen.coroutine
    def shutdown(self):
        """Stops the server configured under this servient."""

        with (yield self._servient_lock.acquire()):
            yield [server.stop() for server in six.itervalues(self._servers)]
            self._stop_catalogue()
            yield self._stop_dnssd()
            self._is_running = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six
import tornado.gen
import tornado.ioloop
from faker import Faker

from tests.utils import run_test_coroutine
from wotpy.protocols.http.client import HTTPClient
from wotpy.protocols.ws.client import WebsocketClient
from wotpy.support import is_coap_supported, is_mqtt_supported
from wotpy.wot.td import ThingDescription


def test_all_protocols_combined(all_protocols_servient):
    """Protocol bindings work as expected when multiple
    servers are combined within the same Servient."""

    exposed_thing = next(all_protocols_servient.exposed_things)
    td = ThingDescription.from_thing(exposed_thing.thing)

    clients = [
        WebsocketClient(),
        HTTPClient()
    ]

    if is_coap_supported():
        from wotpy.protocols.coap.client import CoAPClient
        clients.append(CoAPClient())

    if is_mqtt_supported():
        from tests.protocols.mqtt.broker import is_test_broker_online
        from wotpy.protocols.mqtt.client import MQTTClient
        if is_test_broker_online():
            clients.append(MQTTClient())

    prop_name = next(six.iterkeys(td.properties))

    @tornado.gen.coroutine
    def read_property(the_client):
        prop_value = Faker().sentence()

        curr_value = yield the_client.read_property(td, prop_name)
        assert curr_value != prop_value

        yield exposed_thing.properties[prop_name].write(prop_value)

        curr_value = yield the_client.read_property(td, prop_name)
        assert curr_value == prop_value

    @tornado.gen.coroutine
    def write_property(the_client):
        updated_value = Faker().sentence()

        curr_value = yield exposed_thing.properties[prop_name].read()
        assert curr_value != updated_value

        yield the_client.write_property(td, prop_name, updated_value)

        curr_value = yield exposed_thing.properties[prop_name].read()
        assert curr_value == updated_value

    @tornado.gen.coroutine
    def test_coroutine():
        for client in clients:
            yield read_property(client)
            yield write_property(client)

    run_test_coroutine(test_coroutine)
